"""
The fundamental core classes for the entire package.
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Any

from musify.field import TagField
from musify.printer import AttributePrinter


@dataclass(frozen=True)
class Result(metaclass=ABCMeta):
    """Stores the results of an operation"""
    pass


###########################################################################
## Core item/object types
###########################################################################
class MusifyObject(AttributePrinter, metaclass=ABCMeta):
    """Generic base class for any nameable and taggable object."""

    __slots__ = ("_clean_tags",)

    #: When representing a list of tags as a string, use this value as the separator.
    tag_sep: str = "; "

    @property
    @abstractmethod
    def name(self) -> str:
        """A name for this object"""
        raise NotImplementedError

    @property
    def clean_tags(self) -> dict[TagField, Any]:
        """A map of tags that have been cleaned to use when matching/searching"""
        return self._clean_tags

    def __init__(self):
        self._clean_tags: dict[TagField, Any] = {}

    def __lt__(self, other: MusifyObject):
        return self.name < other.name

    def __gt__(self, other: MusifyObject):
        return self.name > other.name


class MusifyItem(MusifyObject, metaclass=ABCMeta):
    """Generic class for storing an item."""

    __slots__ = ()

    @property
    @abstractmethod
    def uri(self) -> str | None:
        """URI (Uniform Resource Indicator) is the unique identifier for this item."""
        raise NotImplementedError

    @property
    @abstractmethod
    def has_uri(self) -> bool | None:
        """Does this track have a valid associated URI. When None, answer is unknown."""
        raise NotImplementedError

    @abstractmethod
    def __hash__(self):
        raise NotImplementedError

    def __eq__(self, item: MusifyItem):
        """URI attributes equal if both have a URI, names equal otherwise"""
        if self.has_uri and item.has_uri:
            return self.uri == item.uri
        return self.name == item.name

    def __ne__(self, item: MusifyItem):
        return not self.__eq__(item)

    def __getitem__(self, key: str) -> Any:
        """Get the value of a given attribute key"""
        return getattr(self, key)


# noinspection PyPropertyDefinition
class MusifyItemSettable(MusifyItem, metaclass=ABCMeta):
    """Generic class for storing an item that can have select properties modified."""

    __slots__ = ()

    @abstractmethod
    def _uri_getter(self) -> str | None:
        """URI (Uniform Resource Indicator) is the unique identifier for this item."""
        raise NotImplementedError

    @abstractmethod
    def _uri_setter(self, value: str | None):
        """Set both the ``uri`` property and the ``has_uri`` property ."""
        raise NotImplementedError

    uri = property(
        fget=lambda self: self._uri_getter(),
        fset=lambda self, v: self._uri_setter(v),
        doc="URI (Uniform Resource Indicator) is the unique identifier for this item."
    )


class HasLength(metaclass=ABCMeta):
    """Simple protocol for an object which has a length"""

    __slots__ = ()

    @property
    @abstractmethod
    def length(self):
        """Total duration of this object in seconds"""
        raise NotImplementedError



"""
Exceptions relating to remote operations.
"""
from typing import Any

from musify.exception import MusifyError
from musify.libraries.remote.core.types import RemoteIDType, RemoteObjectType


class RemoteError(MusifyError):
    """Exception raised for remote errors"""


class APIError(RemoteError):
    """Exception raised for REST API errors"""


###########################################################################
## Enum errors
###########################################################################
class RemoteIDTypeError(RemoteError):
    """
    Exception raised for remote ID type errors.

    :param message: Explanation of the error.
    :param kind: The ID type related to the error.
    """

    def __init__(self, message: str | None = None, kind: RemoteIDType | None = None, value: Any = None):
        self.kind = kind.name if kind else None
        self.message = message
        formatted = f"{self.kind} | {self.message}" if self.kind else self.message
        formatted += f": {value}" if value else ""
        super().__init__(formatted)


class RemoteObjectTypeError(RemoteError):
    """
    Exception raised for remote object type errors.

    :param message: Explanation of the error.
    :param kind: The item type related to the error.
    """

    def __init__(
            self, message: str | None = None, kind: RemoteObjectType | None = None, value: Any = None):
        self.kind = kind.name if kind else None
        formatted = f"{self.kind} | {message}" if self.kind else message
        formatted += f": {value}" if value else ""
        super().__init__(formatted)



"""
Exceptions relating to processor operations.
"""
from musify.exception import MusifyError


class ProcessorError(MusifyError):
    """Exception raised for errors related to processors."""


class ProcessorLookupError(ProcessorError):
    """Exception raised when processor name given is not valid."""


class ComparerError(ProcessorError):
    """Exception raised for errors related to :py:class:`Comparer` settings."""


class LimiterProcessorError(ProcessorError):
    """Exception raised for errors related to :py:class:`ItemLimiter` logic."""


class MatcherProcessorError(ProcessorError):
    """Exception raised for errors related to :py:class:`ItemMatcher` logic."""


class SorterProcessorError(ProcessorError):
    """Exception raised for errors related to :py:class:`ItemSorter` logic."""


class TimeMapperError(ProcessorError):
    """Exception raised for errors related to :py:class:`TimeMapper` logic."""


###########################################################################
## Filter errors
###########################################################################
class FilterError(ProcessorError):
    """Exception raised for errors related to :py:class:`Filter` logic."""



"""
An implementation of :py:class:`LocalLibrary` for the MusicBee library manager.
Reads library/settings files from MusicBee to load and enrich playlist/track etc. data.
"""
import hashlib
import os
import re
from collections.abc import Iterable, Mapping, Sequence, Collection, Iterator
from datetime import datetime
from pathlib import Path
from typing import Any
from urllib.parse import quote, unquote

from musify.file.base import File
from musify.file.exception import FileDoesNotExistError, UnexpectedPathError
from musify.file.path_mapper import PathMapper, PathStemMapper
from musify.libraries.local.exception import MusicBeeIDError, XMLReaderError
from musify.libraries.local.library.library import LocalLibrary
from musify.libraries.local.playlist import LocalPlaylist
from musify.libraries.local.track import LocalTrack
from musify.libraries.remote.core.wrangle import RemoteDataWrangler
from musify.processors.base import Filter
from musify.types import Number
from musify.utils import to_collection, required_modules_installed

try:
    import xmltodict
    from lxml import etree
    # noinspection PyProtectedMember
    from lxml.etree import _Element as Element
except ImportError:
    xmltodict = None
    etree = None

    from typing import Never
    Element = Never

REQUIRED_MODULES = [xmltodict, etree]


class MusicBee(LocalLibrary, File):
    """
    Represents a local MusicBee library, providing various methods for manipulating
    tracks and playlists across an entire local library collection.

    :param musicbee_folder: The absolute path of the musicbee folder containing settings and library files.
    :param playlist_filter: An optional :py:class:`Filter` to apply or collection of playlist names to include when
        loading playlists. Playlist names will be passed to this filter to limit which playlists are loaded.
    :param path_mapper: Optionally, provide a :py:class:`PathMapper` for paths stored in the playlist files.
        Useful if the playlist files contain relative paths and/or paths for other systems that need to be
        mapped to absolute, system-specific paths to be loaded and back again when saved.
    :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object for processing URIs on tracks.
        If given, the wrangler can be used when calling __get_item__ to get an item from the collection from its URI.
        The wrangler is also used when loading tracks to allow them to process URI tags.
        For more info on this, see :py:class:`LocalTrack`.
    :param name: A name to assign to this library.
    """

    __slots__ = (
        "musicbee_folder",
        "library_xml",
        "_library_xml_path",
        "_library_xml_parser",
        "settings_xml",
        "_settings_xml_path",
        "_settings_xml_parser",
    )

    valid_extensions = frozenset({".xml"})
    #: The path of the MusicBee library file relative the ``musicbee_folder`` provided on initialisation.
    xml_library_path = Path("iTunes Music Library.xml")
    #: A list of keys for the XML library that need to be processed as system paths.
    xml_library_path_keys = {"Location", "Music Folder"}
    #: The path of the MusicBee settings file relative the ``musicbee_folder`` provided on initialisation.
    xml_settings_path = Path("MusicBeeLibrarySettings.ini")
    #: The path to the playlists folder relative the ``musicbee_folder`` provided on initialisation.
    playlists_path = Path("Playlists")

    @property
    def path(self) -> Path:
        return self._library_xml_path

    def __init__(
            self,
            musicbee_folder: str | Path,
            playlist_filter: Collection[str] | Filter[str] = (),
            path_mapper: PathMapper = PathMapper(),
            remote_wrangler: RemoteDataWrangler = None,
            name: str = None,
    ):
        required_modules_installed(REQUIRED_MODULES, self)

        #: The absolute path of the musicbee folder containing settings and library files.
        self.musicbee_folder = Path(musicbee_folder)
        if not self.musicbee_folder.is_dir():
            raise UnexpectedPathError(self.musicbee_folder, "Given MusicBee folder must be a directory")

        self._library_xml_path: Path = self.musicbee_folder.joinpath(self.xml_library_path)
        if not self._library_xml_path.is_file():
            raise FileDoesNotExistError(self._library_xml_path, "Cannot find MusicBee library file at this path")

        try:
            self._library_xml_parser = XMLLibraryParser(self._library_xml_path, path_keys=self.xml_library_path_keys)
            #: A map representation of the loaded XML library data
            self.library_xml: dict[str, Any] = self._library_xml_parser.parse()
        except (etree.XMLSyntaxError, XMLReaderError) as exc:
            raise XMLReaderError(
                f"Could not read from library file at {self._library_xml_path}. {exc}"
            ) from exc

        self._settings_xml_path: Path = self.musicbee_folder.joinpath(self.xml_settings_path)
        if not self._settings_xml_path.is_file():
            raise FileDoesNotExistError(self._settings_xml_path, "Cannot find MusicBee settings file at this path")

        try:
            with open(self._settings_xml_path, "r", encoding="utf-8") as f:
                #: A map representation of the loaded XML settings data
                self.settings_xml: dict[str, Any] = xmltodict.parse(f.read())["ApplicationSettings"]
        except (etree.XMLSyntaxError, XMLReaderError) as exc:
            raise XMLReaderError(
                f"Could not read from settings file at {self._settings_xml_path}. {exc}"
            ) from exc

        library_folders = []
        for path in to_collection(self.settings_xml.get("OrganisationMonitoredFolders", {}).get("string")):
            library_folders.append(path_mapper.map(path))

        super().__init__(
            library_folders=library_folders,
            playlist_folder=self.musicbee_folder.joinpath(self.playlists_path),
            playlist_filter=playlist_filter,
            path_mapper=path_mapper,
            remote_wrangler=remote_wrangler,
            name=name,
        )

    def _get_track_from_xml_path(
            self, track_xml: dict[str, Any], track_map: dict[str, LocalTrack]
    ) -> LocalTrack | None:
        if track_xml["Track Type"] != "File":
            return

        path = track_xml["Location"]
        prefixes = {*map(str, self.library_folders), self.library_xml["Music Folder"]}
        if isinstance(self.path_mapper, PathStemMapper):
            prefixes.update(self.path_mapper.stem_map.keys())

        for prefix in prefixes:
            track = track_map.get(path.removeprefix(prefix).casefold())
            if track is not None:
                return track

        self.errors.append(path)

    def _map_track_to_xml(self) -> dict[LocalTrack, dict[str, Any]]:
        # need to remove library folders to allow match to be os agnostic
        track_map = {
            str(track.path).removeprefix(str(folder)).casefold(): track
            for folder in self.library_folders for track in self.tracks
        }
        track_xml_map: dict[LocalTrack, dict[str, Any]] = {}

        for track_xml in self.library_xml["Tracks"].values():
            track = self._get_track_from_xml_path(track_xml=track_xml, track_map=track_map)
            if track is None:
                continue

            track_xml_map[track] = track_xml

        return track_xml_map

    async def load_tracks(self) -> None:
        await super().load_tracks()
        self.logger.debug(f"Enrich {self.name} tracks: START")

        for track, track_xml in self._map_track_to_xml().items():
            track.rating = int(track_xml.get("Rating")) if track_xml.get("Rating") is not None else None
            track.date_added = track_xml.get("Date Added")
            track.last_played = track_xml.get("Play Date UTC")
            track.play_count = track_xml.get("Play Count", 0)

        self._log_errors("Could not find a loaded track for these paths from the MusicBee library file")
        self.logger.debug(f"Enrich {self.name} tracks: DONE\n")

    async def save(self, dry_run: bool = True, *_, **__) -> dict[str, Any]:
        """
        Generate and save the XML library file for this MusicBee library.

        :param dry_run: Run function, but do not modify the file on the disk.
        :return: Map representation of the saved XML file.
        """
        self.logger.debug(f"Save {self.name} library file: START")

        track_id_map: dict[LocalTrack, tuple[int, str]] = {
            track: (track_xml["Track ID"], track_xml["Persistent ID"])
            for track, track_xml in self._map_track_to_xml().items()
        }

        self._log_errors("Could not find a loaded track for these paths from the MusicBee library file")

        tracks: dict[int, dict[str, Any]] = {}
        max_track_id = max(id_ for id_, _ in track_id_map.values()) if track_id_map else 0
        for i, track in enumerate(self.tracks, max(1, max_track_id + 1)):
            track_id, persistent_id = track_id_map.get(track, [i, None])
            tracks[track_id] = self.track_to_xml(track, track_id=track_id, persistent_id=persistent_id)
            track_id_map[track] = (tracks[track_id]["Track ID"], tracks[track_id]["Persistent ID"])

        playlist_id_map = {
            playlist_xml["Name"]: (playlist_xml["Playlist ID"], playlist_xml["Playlist Persistent ID"])
            for playlist_xml in self.library_xml["Playlists"]
        }

        playlists: list[dict[str, Any]] = []
        max_playlist_id = max(id_ for id_, _ in playlist_id_map.values()) if playlist_id_map else 0
        for i, (name, playlist) in enumerate(self.playlists.items(), max_playlist_id + 1):
            playlist_id, persistent_id = playlist_id_map.get(name, [i, None])
            playlist = self.playlist_to_xml(
                playlist, tracks=track_id_map, playlist_id=playlist_id, persistent_id=persistent_id
            )
            playlists.append(playlist)

        xml = {
            "Major Version": self.library_xml.get("Major Version", "1"),
            "Minor Version": self.library_xml.get("Minor Version", "1"),
            "Application Version": self.library_xml.get("Application Version", "1"),
            "Music Folder": str(self.musicbee_folder),
            "Library Persistent ID":
                self.library_xml.get("Library Persistent ID", self.generate_persistent_id(self.musicbee_folder)),
            "Tracks": dict(sorted(((id_, track) for id_, track in tracks.items()), key=lambda x: x[0])),
            "Playlists": sorted(playlists, key=lambda x: x["Playlist ID"]),
        }

        self._library_xml_parser.unparse(data=xml, dry_run=dry_run)
        self.library_xml = xml

        self.logger.debug(f"Save {self.name} library file: DONE")
        return self.library_xml

    @staticmethod
    def generate_persistent_id(value: str | Path | None = None, id_: str | None = None) -> str:
        """
        Generates a valid persistent ID from a given ``value``
        or validates a given persistent ID as given by ``id_``,

        :param id_: A persistent ID to validate
        :param value: A value to generate a persistent ID from.
        :return: The valid persistent ID.
        :raise MusicBeeIDError: When no ``id_`` and no ``value`` is given, or the given ``id_`` is invalid.
        """
        if not value and not id_:
            raise MusicBeeIDError(
                "You must provide either a persistent ID to validate or a value to generate a persistent ID from."
            )

        id_ = id_ or hashlib.sha256(str(value).encode("utf-8")).hexdigest()[:16]
        if len(id_) > 16:
            raise MusicBeeIDError(f"Persistent ID is >16-characters in length (length={len(id_)}): {id_}")
        return id_.upper()

    @classmethod
    def track_to_xml(cls, track: LocalTrack, track_id: int, persistent_id: str | None = None) -> dict[str, Any]:
        """
        Convert a local track into a dict representation of XML data for a MusicBee library file

        :param track: The track to convert.
        :param track_id: An incremental ID to assign to the track.
        :param persistent_id: An 16-character unique ID to assign to the track.
        :return: A dict representation of XML data for the given track.
        :raise MusicBeeIDError: When the given ``persistent_id`` is invalid.
        """
        genres = {}
        if track.genres and len(track.genres) == 1:
            genres = {"Genre": track.genres[0]}
        elif track.genres:
            genres = {f"Genre{i}": genre for i, genre in enumerate(track.genres, 1)}

        data = {
            "Track ID": track_id,
            "Persistent ID": cls.generate_persistent_id(id_=persistent_id, value=track.path),
            "Name": track.title,
            "Artist": track.artist,
            "Album": track.album,
            "Album Artist": track.album_artist,
            "Track Number": track.track_number,
            "Track Count": track.track_total,
        } | genres | {
            "Year": track.year,
            "BPM": track.bpm,
            "Disc Number": track.disc_number,
            "Disc Count": track.disc_total,
            "Compilation": track.compilation,
            "Comments": track.tag_sep.join(track.comments) if track.comments else None,
            "Total Time": int(track.length * 1000),  # in milliseconds
            "Rating": track.rating,
            # "Composer": track.composer,  # currently not supported by this program
            # "Conductor": track.conductor,  # currently not supported by this program
            # "Publisher": track.publisher,  # currently not supported by this program
            # "Encoder": track.encoder,  # currently not supported by this program
            "Size": track.size,
            "Kind": track.type,
            # "": track.channels,  # unknown MusicBee mapping
            "Bit Rate": int(track.bit_rate),
            # "": track.bit_depth,  # unknown MusicBee mapping
            "Sample Rate": int(track.sample_rate * 1000),  # in Hz
            "Date Modified": track.date_modified,
            "Date Added": track.date_added,
            "Play Date UTC": track.last_played,
            "Play Count": track.play_count,
            "Track Type": "File",  # can also be 'URL' for streams
            "Location": str(track.path),
        }

        return dict(filter(lambda item: item[1] is not None, data.items()))

    @classmethod
    def playlist_to_xml(
            cls,
            playlist: LocalPlaylist,
            tracks: Mapping[LocalTrack, int | Sequence[int, str]] | list[LocalTrack],
            playlist_id: int,
            persistent_id: str | None = None,
    ) -> dict[str, Any]:
        """
        Convert a local playlist into a dict representation of XML data for a MusicBee library file.

        :param playlist: The playlist to convert.
        :param tracks: A map of all available tracks in the library to their track IDs, or a simple list of tracks.
            If a map is given and the values are :py:class:`Sequence`, take the first index as the Track ID.
            If a list is given, the function will use the track's index as the Track ID.
        :param playlist_id: An incremental ID to assign to the playlist.
        :param persistent_id: An 16-character unique ID to assign to the playlist.
        :return: A dict representation of XML data for the given playlist.
        :raise MusicBeeIDError: When the given ``persistent_id`` is invalid.
        """
        items: list[dict[str, int]] = []
        for track in playlist:
            if isinstance(tracks, dict):
                result = tracks[track]
                items.append({"Track ID": result if isinstance(result, int) else result[0]})
            else:
                items.append({"Track ID":  tracks.index(track)})

        data = {
            "Playlist ID": playlist_id,
            "Playlist Persistent ID": cls.generate_persistent_id(id_=persistent_id, value=playlist.path),
            "All Items": True,  # unknown what this does, what happens if 'False'?
            "Name": playlist.name,
            "Description": playlist.description,
            "Playlist Items": items,
        }

        return dict(filter(lambda item: item[1] is not None, data.items()))


# noinspection PyProtectedMember
class XMLLibraryParser:
    """
    Parses MusicBee XML files to and from iTunes style XML.

    :param path: Path to the XML file.
    :param path_keys: A list of keys in the XML file that need to be processed as system paths.
    """

    __slots__ = ("path", "path_keys", "_iterparse")

    #: The string representation of the timestamp format when parsing.
    timestamp_format = "%Y-%m-%dT%H:%M:%SZ"

    def __init__(self, path: str | Path, path_keys: Iterable[str] | None = None):
        required_modules_installed(REQUIRED_MODULES, self)

        #: Path to the XML file.
        self.path: Path = Path(path)
        #: A list of keys in the XML file that need to be processed as system paths.
        self.path_keys: frozenset[str] = frozenset(path_keys) if path_keys else frozenset()
        #: Stores the iterparse operator for parsing XML file
        self._iterparse: etree.iterparse | None = None

    @classmethod
    def to_xml_timestamp(cls, timestamp: datetime | None) -> str | None:
        """Convert timestamp string as found in the MusicBee XML library file to a ``datetime`` object"""
        if timestamp:
            return timestamp.strftime(cls.timestamp_format)

    @classmethod
    def from_xml_timestamp(cls, timestamp_str: str | None) -> datetime | None:
        """Convert timestamp string as found in the MusicBee XML library file to a ``datetime`` object"""
        if timestamp_str:
            return datetime.strptime(timestamp_str, cls.timestamp_format)

    @staticmethod
    def to_xml_path(path: str | Path) -> str:
        """Convert a standard system path to a file path as found in the MusicBee XML library file"""
        return f"file://localhost/{quote(str(path).replace('\\', '/'), safe=':/!(),;@[]+')}"\
            .replace("%26", "&#38;")\
            .replace("%27", "&#39;")

    @staticmethod
    def from_xml_path(path: str | Path) -> str:
        """Clean the file paths as found in the MusicBee XML library file to a standard system path"""
        return os.path.normpath(unquote(str(path).removeprefix("file://localhost/")))

    def _iter_elements(self) -> Iterator[Element]:
        for event, element in self._iterparse:
            yield element

    def _parse_value(self, value: Any, tag: str, parent: str | None = None):
        if tag == 'string':
            if parent in self.path_keys:
                return self.from_xml_path(value)
            else:
                return value
        elif tag == 'integer':
            try:
                return int(value) if "." not in value else float(value)
            except ValueError:
                return value
        elif tag == 'date':
            return self.from_xml_timestamp(value)
        elif tag in ['true', 'false']:
            return tag == 'true'

    def _parse_element(self, element: Element | None = None) -> Any:
        elem = next(self._iter_elements())
        peek = element.getnext() if element is not None else None

        if elem.tag in ['string', 'integer', 'date', 'true', 'false']:
            return self._parse_value(
                value=elem.text, tag=elem.tag, parent=element.text if element is not None else None
            )
        elif peek is not None and peek.tag == "dict":
            next_elem = next(self._iter_elements())
            value = self._parse_value(value=next_elem.text, tag=next_elem.tag, parent=elem.text)
            return {elem.text: value} | self._parse_dict()
        elif peek is not None and peek.tag == "array":
            return self._parse_array(elem)
        elif peek is not None:
            raise XMLReaderError(f"Unrecognised element: {element.tag}, {element.text}, {peek.tag}, {peek.text}")
        elif element is not None:
            raise XMLReaderError(f"Unrecognised element: {element.tag}, {element.text}")
        else:
            raise XMLReaderError(f"Unrecognised element: {elem.tag}, {elem.text}")

    def _parse_array(self, element: Element | None = None) -> list[Any]:
        array = []

        if element is not None and element.tag == "array" and element.text is None:
            return array  # array is empty, skip processing

        for elem in self._iter_elements():
            if elem is None or elem.tag == "array":
                break

            peek = elem.getnext()
            if elem.tag == "key":
                next_elem = next(self._iter_elements())
                value = self._parse_value(value=next_elem.text, tag=next_elem.tag, parent=elem.text)
                array.append({elem.text: value} | self._parse_dict())
            elif peek is None and element is not None:
                value = self._parse_value(value=elem.text, tag=elem.tag, parent=element.text)
                array.append({element.text: value} | self._parse_dict())
            elif elem.tag != "plist":
                array.append(self._parse_element(elem))

        return array

    def _parse_dict(self) -> dict[str, Any]:
        record = {}

        for elem in self._iter_elements():
            if elem is None or elem.tag == "dict":
                break

            peek = elem.getnext()
            if peek is not None and peek.tag == "dict":
                record[elem.text] = self._parse_dict()
            else:
                record[elem.text] = self._parse_element(elem)

        return record

    def parse(self) -> dict[str, Any]:
        """Parse the XML file from the currently stored ``path`` to a dictionary"""
        root_name = etree.parse(self.path).docinfo.root_name
        self._iterparse = etree.iterparse(self.path)
        results = {}

        for element in self._iter_elements():
            peek = element.getnext()
            if peek is None:
                break

            if element.tag == root_name:
                continue
            elif element.tag == "key":
                key = element.text
                if peek.tag == "dict":
                    results[key] = self._parse_dict()
                elif peek.tag == "array":
                    results[key] = self._parse_array()
                else:
                    results[key] = self._parse_element(element)
            else:
                raise NotImplementedError

        # close the iterator
        for _ in self._iterparse:
            pass
        self._iterparse = None

        return results

    def _unparse_dict(self, element: Element, data: Mapping[str, Any]):
        sub_element: etree._Element = etree.SubElement(element, "dict")
        for key, value in data.items():
            etree.SubElement(sub_element, "key").text = str(key)

            if isinstance(value, bool):
                etree.SubElement(sub_element, str(value).lower())
            elif isinstance(value, str | Path):
                value = str(value)
                if key in self.path_keys:
                    etree.SubElement(sub_element, "string").text = self.to_xml_path(value)
                else:
                    etree.SubElement(sub_element, "string").text = str(value)
            elif isinstance(value, Number):
                etree.SubElement(sub_element, "integer").text = str(value)
            elif isinstance(value, datetime):
                etree.SubElement(sub_element, "date").text = self.to_xml_timestamp(value)
            elif isinstance(value, Mapping):
                self._unparse_dict(element=sub_element, data=value)
            elif isinstance(value, Sequence):
                array_element: etree.Element = etree.SubElement(sub_element, "array")
                for item in value:
                    self._unparse_dict(element=array_element, data=item)
            else:
                raise XMLReaderError(f"Unexpected value type: {value} ({type(value)})")

    def unparse(self, data: Mapping[str, Any], dry_run: bool = True) -> None:
        """
        Un-parse a map of XML data to XML and save to file.

        :param data: Map of XML data to export.
        :param dry_run: Run function, but do not modify the file on the disk.
        """
        et: etree._ElementTree = etree.parse(self.path)
        parsed: dict[str, Any] = xmltodict.parse(etree.tostring(et.getroot(), encoding='utf-8', method='xml'))

        root: etree.Element = etree.Element(et.docinfo.root_name)
        root.set("version", parsed[et.docinfo.root_name]["@version"])

        self._unparse_dict(element=root, data=data)
        etree.indent(root, space="\t", level=0)

        # convert to string and apply formatting to ensure output string is expected format
        output: str = etree.tostring(
            root, xml_declaration=True, encoding=et.docinfo.encoding, doctype=et.docinfo.doctype
        ).decode(et.docinfo.encoding)
        output = re.sub(r"</key>\n\s+<(string|integer|date|true|false)", r"</key><\1", output)
        output = re.sub("\n\t", "\n", output)
        output = output.replace("'", '"')

        if not dry_run:
            with open(self.path, "w", encoding="utf-8") as file:
                file.write(output.rstrip('\n') + '\n')



"""
The FLAC implementation of a :py:class:`LocalTrack`.
"""
from io import BytesIO
from typing import Any

import mutagen
import mutagen.flac
import mutagen.id3

from musify.field import TagMap
from musify.file.image import open_image, get_image_bytes
# noinspection PyProtectedMember
from musify.libraries.local.track._tags import TagReader, TagWriter
from musify.libraries.local.track.field import LocalTrackField
from musify.libraries.local.track.track import LocalTrack

try:
    from PIL import Image
except ImportError:
    Image = None


class _FLACTagReader(TagReader[mutagen.flac.FLAC]):

    __slots__ = ()

    def read_images(self):
        if Image is None:
            return

        values = self.file.pictures
        return [Image.open(BytesIO(value.data)) for value in values] if len(values) > 0 else None

    def check_for_images(self) -> bool:
        return len(self.file.pictures) > 0


class _FLACTagWriter(TagWriter[mutagen.flac.FLAC]):

    __slots__ = ()

    def _write_tag(self, tag_id: str | None, tag_value: Any, dry_run: bool = True) -> bool:
        if not dry_run:
            if isinstance(tag_value, (list, set, tuple)):
                self.file[tag_id] = list(map(str, tag_value))
            else:
                self.file[tag_id] = str(tag_value)
        return True

    def _write_images(self, track: LocalTrack, dry_run: bool = True) -> bool:
        updated = False
        for image_kind, image_link in track.image_links.items():
            image = open_image(image_link)
            image_kind_attr = image_kind.upper().replace(" ", "_")

            picture = mutagen.flac.Picture()
            picture.type = getattr(mutagen.id3.PictureType, image_kind_attr)
            picture.mime = Image.MIME[image.format]
            picture.data = get_image_bytes(image)

            if not dry_run:
                # clear all images, adding back those that don't match the new image's type
                pictures_current = self.file.pictures.copy()
                self.file.clear_pictures()
                for pic in pictures_current:
                    if pic.type != picture.type:
                        self.file.add_picture(pic)

                self.file.add_picture(picture)

            image.close()
            track.has_image = True
            updated = True

        return updated

    def _clear_tag(self, tag_name: str, dry_run: bool = True) -> bool:
        if tag_name == LocalTrackField.IMAGES.name.lower():
            self.file.clear_pictures()
            return True

        removed = False

        tag_ids = self.tag_map[tag_name]
        if tag_ids is None or len(tag_ids) is None:
            return removed

        for tag_id in tag_ids:
            if tag_id in self.file and self.file[tag_id]:
                if not dry_run:
                    del self.file[tag_id]
                removed = True

        return removed


class FLAC(LocalTrack[mutagen.flac.FLAC, _FLACTagReader, _FLACTagWriter]):

    __slots__ = ()

    valid_extensions = frozenset({".flac"})

    # noinspection SpellCheckingInspection
    #: Map of human-friendly tag name to ID3 tag ids for a given file type
    tag_map = TagMap(
        title=["title"],
        artist=["artist"],
        album=["album"],
        album_artist=["albumartist"],
        track_number=["tracknumber", "track"],
        track_total=["tracktotal"],
        genres=["genre"],
        date=["date", "year"],
        year=["year"],
        bpm=["bpm"],
        key=["initialkey"],
        disc_number=["discnumber"],
        disc_total=["disctotal"],
        compilation=["compilation"],
        comments=["comment", "description"],
    )

    def _create_reader(self, file: mutagen.flac.FLAC):
        return _FLACTagReader(file, tag_map=self.tag_map, remote_wrangler=self._remote_wrangler)

    def _create_writer(self, file: mutagen.flac.FLAC):
        return _FLACTagWriter(file, tag_map=self.tag_map, remote_wrangler=self._remote_wrangler)



"""
Core abstract classes for the :py:mod:`Remote` module.

These define the foundations of any remote object or item.
"""
from abc import ABCMeta, abstractmethod
from typing import Any, Self

from musify.base import MusifyItem
from musify.libraries.remote.core import RemoteResponse
from musify.libraries.remote.core.api import RemoteAPI
from musify.libraries.remote.core.exception import APIError
from musify.libraries.remote.core.types import APIInputValueSingle, RemoteObjectType


class RemoteObject[T: (RemoteAPI | None)](RemoteResponse, metaclass=ABCMeta):
    """
    Generic base class for remote objects. Extracts key data from a remote API JSON response.

    :param response: The remote API JSON response
    :param api: The instantiated and authorised API object for this source type.
    """

    __slots__ = ("_response", "api")
    __attributes_ignore__ = ("response", "api")

    @property
    @abstractmethod
    def uri(self) -> str:
        """URI (Uniform Resource Indicator) is the unique identifier for this item/collection."""
        raise NotImplementedError

    @property
    @abstractmethod
    def has_uri(self) -> bool:
        """Does this item/collection have a valid URI that is not a local URI."""
        raise NotImplementedError

    @property
    def response(self) -> dict[str, Any]:
        """The API response for this object"""
        return self._response

    # noinspection PyPropertyDefinition,PyMethodParameters
    @property
    @abstractmethod
    def kind(cls) -> RemoteObjectType:
        raise NotImplementedError

    def __init__(self, response: dict[str, Any], api: T = None, skip_checks: bool = False):
        super().__init__()
        self._response = response

        #: The :py:class:`RemoteAPI` to call when reloading
        self.api = api

        self._check_type()
        self.refresh(skip_checks=skip_checks)

    async def __aenter__(self) -> Self:
        await self.api.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.api.__aexit__(exc_type, exc_val, exc_tb)

    @abstractmethod
    def _check_type(self) -> None:
        """
        Checks the given response is compatible with this object type, raises an exception if not.

        :raise RemoteObjectTypeError: When the response type is not compatible with this object.
        """
        raise NotImplementedError

    def _check_for_api(self) -> None:
        """
        Checks the API has been set on the instance, raises an exception if not.

        :raise APIError: When the API has not been set for this instance.
        """
        if self.api is None:
            raise APIError("API is not set. Assign an API to this instance first.")

    @classmethod
    @abstractmethod
    async def load(
            cls, value: APIInputValueSingle[RemoteResponse], api: RemoteAPI, *args, **kwargs
    ) -> Self:
        """
        Generate a new object of this class,
        calling all required endpoints to get a complete set of data for this item type.

        ``value`` may be:
            * A string representing a URL/URI/ID.
            * A remote API JSON response for a collection with a valid ID value under an ``id`` key.
            * An object of the same type as this collection.
              The remote API JSON response will be used to load a new object.

        :param value: The value representing some remote object. See description for allowed value types.
        :param api: An authorised API object to load the object from.
        """
        raise NotImplementedError

    @abstractmethod
    async def reload(self, *args, **kwargs) -> None:
        """
        Reload this object from the API, calling all required endpoints
        to get a complete set of data for this item type.
        """
        raise NotImplementedError

    def __hash__(self):
        """Uniqueness of a remote object is its URI"""
        return hash(self.uri)


class RemoteItem(RemoteObject, MusifyItem, metaclass=ABCMeta):
    """Generic base class for remote items. Extracts key data from a remote API JSON response."""

    __attributes_classes__ = (RemoteObject, MusifyItem)



"""
All type hints to use throughout the module.
"""
from collections.abc import Mapping, MutableMapping
from typing import Any

from aiorequestful.types import URLInput
from yarl import URL

from musify.libraries.remote.core import RemoteResponse
from musify.types import UnitMutableSequence, UnitSequence, MusifyEnum

type APIInputValueSingle[T: RemoteResponse] = URLInput | Mapping[str, Any] | T
type APIInputValueMulti[T: RemoteResponse] = (
        UnitSequence[str] |
        UnitMutableSequence[URL] |
        UnitMutableSequence[MutableMapping[str, Any]] |
        UnitSequence[T]
)


class RemoteIDType(MusifyEnum):
    """Represents remote ID types"""
    ALL: int = 0

    #: Value is the expected length of ID string
    ID: int = 22
    #: Value is the expected number of chunks in a URI
    URI: int = 3
    URL: int = 1
    URL_EXT: int = 2


class RemoteObjectType(MusifyEnum):
    """Represents remote object types"""
    ALL = 0
    PLAYLIST = 1
    TRACK = 2
    ALBUM = 3
    ARTIST = 4
    USER = 5
    SHOW = 6
    EPISODE = 7
    AUDIOBOOK = 8
    CHAPTER = 9



"""
Processor that sorts the given collection of items based on given configuration.
"""
import random
from collections.abc import Callable, Mapping, MutableMapping, Sequence, Iterable
from copy import copy
from datetime import datetime
from random import shuffle
from typing import Any

from musify.base import MusifyItem
from musify.field import Field
from musify.processors.base import Processor
from musify.processors.exception import SorterProcessorError
from musify.types import UnitSequence, UnitIterable, Number, MusifyEnum
from musify.utils import flatten_nested, strip_ignore_words, to_collection, limit_value


class ShuffleMode(MusifyEnum):
    """Represents the possible shuffle modes to use when shuffling items using :py:class:`ItemSorter`."""
    RANDOM = 0
    HIGHER_RATING = 1
    RECENT_ADDED = 2
    DIFFERENT_ARTIST = 3


class ItemSorter(Processor):
    """
    Sort items in-place based on given conditions.

    ``fields`` may be:
        * List of tags/properties to sort by.
        * Map of ``{<tag/property>: <reversed>}``. If reversed is true, sort the ``tag/property`` in reverse.

    When ``shuffle_mode`` == ``HIGHER_RATING`` or ``RECENT_ADDED``:
        * A ``shuffle_weight`` of 0 will sort the tracks in order according to the desired ``shuffle_mode``.
        * A positive ``shuffle_weight`` shuffles according to the desired ``shuffle_mode``.
          The ``shuffle_weight`` will determine how much randomness is applied to lower ranking items.
        * A negative ``shuffle_weight`` works as above but reverses the final sort order.

    When ``shuffle_mode`` == ``DIFFERENT_ARTIST``:
        * A ``shuffle_weight`` of 1 will group the tracks by artist, shuffling artists randomly.
        * A ``shuffle_weight`` of -1 will shuffle the items randomly.

    :param fields: Fields to sort by. If defined, this value will always take priority over any shuffle settings
        i.e. shuffle settings will be ignored.
    :param shuffle_mode: The mode to use for shuffling. Only used when no ``fields`` are given.
        WARNING: Currently only ``RANDOM`` shuffle mode has been implemented.
        Any other given value will default to ``RANDOM`` shuffling.
    :param shuffle_weight: The weights (between -1 and 1) to apply to certain shuffling modes.
        This value will automatically be limited to within the valid range -1 and 1.
        Only used when no ``fields`` are given and shuffle_mode is not None or ``RANDOM``.
    """

    __slots__ = ("sort_fields", "shuffle_mode", "shuffle_weight")

    @classmethod
    def sort_by_field(cls, items: list[MusifyItem], field: Field | None = None, reverse: bool = False) -> None:
        """
        Sort items by the values of a given field.

        :param items: List of items to sort
        :param field: Tag or property to sort on. If None and reverse is True, reverse the order of the list.
        :param reverse: If true, reverse the order of the sort.
        """
        if field is None:
            if reverse:
                items.reverse()
            return

        tag_name = field.map(field)[0].name.lower()

        # attempt to find an example value to determine the value type for this sort
        example_value = None
        for item in items:
            example_value = item[tag_name]
            if example_value is not None:
                break
        if example_value is None:
            # if no example value found, all values are None and so no sort can happen safely. Skip
            return

        # get sort key based on value type
        if isinstance(example_value, datetime):  # key converts datetime to floats
            def sort_key(t: MusifyItem) -> float:
                """Get the sort key for timestamp tags from the given ``t``"""
                value = t[tag_name]
                return value.timestamp() if value is not None else 0.0
        elif isinstance(example_value, str):  # key strips ignore words from string
            sort_key: Callable[[MusifyItem], (bool, str)] = lambda t: strip_ignore_words(t[tag_name])
        else:
            sort_key: Callable[[MusifyItem], object] = lambda t: t[tag_name] if t[tag_name] else 0

        items.sort(key=sort_key, reverse=reverse)

    @classmethod
    def group_by_field[T: MusifyItem](cls, items: UnitIterable[T], field: Field | None = None) -> dict[Any, list[T]]:
        """
        Group items by the values of a given field.

        :param items: List of items to sort.
        :param field: Tag or property to group by. None returns map of ``{None: <items>}``.
        :return: Map of grouped items.
        """
        if field is None:  # group by None
            return {None: to_collection(items, list)}

        tag_name = field.map(field)[0].name.lower()

        def group(v: Any) -> None:
            """Group items by the given value ``v``"""
            if grouped.get(v) is None:
                grouped[v] = []
            grouped[v].append(item)

        grouped: dict[Any | None, list[T]] = {}
        for item in items:  # produce map of grouped values
            value = to_collection(item[tag_name])
            if isinstance(value, Iterable):
                for val in value:
                    group(val)
            else:
                group(value)

        return grouped

    def __init__(
            self,
            fields: UnitSequence[Field | None] | Mapping[Field | None, bool] = (),
            shuffle_mode: ShuffleMode | None = None,
            shuffle_weight: float = 0.0
    ):
        fields = to_collection(fields, list) if isinstance(fields, Field) else fields
        self.sort_fields: dict[Field | None, bool] = {field: False for field in fields} \
            if isinstance(fields, Sequence) else fields

        self.shuffle_mode = shuffle_mode
        self.shuffle_weight = limit_value(shuffle_weight, floor=-1, ceil=1)

    def __call__(self, *args, **kwargs) -> None:
        return self.sort(*args, **kwargs)

    def sort(self, items: list[MusifyItem]) -> None:
        """Sorts a list of ``items`` in-place."""
        if len(items) == 0:
            return

        if self.sort_fields:
            items_nested = self._sort_by_fields({None: items}, fields=self.sort_fields)
            items.clear()
            items.extend(flatten_nested(items_nested))
        elif self.shuffle_mode == ShuffleMode.RANDOM:
            shuffle(items)
        elif self.shuffle_mode == ShuffleMode.HIGHER_RATING:
            self._shuffle_on_rating(items)
        elif self.shuffle_mode == ShuffleMode.RECENT_ADDED:
            self._shuffle_on_date_added(items)
        elif self.shuffle_mode == ShuffleMode.DIFFERENT_ARTIST:
            self._shuffle_on_artist(items)

    def _get_weighted_shuffle_value(self, value: Number, max_value: Number) -> float:
        weight_factor = random.uniform(-1, 1) * self.shuffle_weight
        return abs(value - weight_factor * (value - max_value))

    # noinspection PyUnresolvedReferences
    def _shuffle_on_rating(self, items: list[MusifyItem]) -> None:
        if not all(hasattr(item, "rating") for item in items):
            raise SorterProcessorError(
                "Cannot shuffle sort on rating as the given items do not all have a 'rating' property"
            )

        max_value: float = max(item.rating for item in items)
        items.sort(
            key=lambda item: self._get_weighted_shuffle_value(item.rating, max_value),
            reverse=self.shuffle_weight >= 0
        )

    # noinspection PyUnresolvedReferences
    def _shuffle_on_date_added(self, items: list[MusifyItem]) -> None:
        if not all(hasattr(item, "date_added") for item in items):
            raise SorterProcessorError(
                "Cannot shuffle sort on date added as the given items do not all have a 'date_added' property"
            )

        max_value: float = max(item.date_added.timestamp() for item in items)
        items.sort(
            key=lambda item: self._get_weighted_shuffle_value(item.date_added.timestamp(), max_value),
            reverse=self.shuffle_weight >= 0
        )

    # noinspection PyUnresolvedReferences
    def _shuffle_on_artist(self, items: list[MusifyItem]) -> None:
        if not all(hasattr(item, "artist") for item in items):
            raise SorterProcessorError(
                "Cannot shuffle sort on artist as the given items do not all have an 'artist' property"
            )

        shuffle_weight = (self.shuffle_weight + 1) / 2
        artists: list[str] = list({item.artist for item in items})
        shuffle(artists)

        def sort_key(artist: str) -> int:
            """Get sort key for a given ``artist``"""
            return artists.index(artist) if random.random() <= shuffle_weight else random.randrange(0, len(artists))

        shuffle(items)
        items.sort(key=lambda item: sort_key(item.artist))

    @classmethod
    def _sort_by_fields(
            cls, items_grouped: MutableMapping, fields: MutableMapping[Field | None, bool]
    ) -> MutableMapping:
        """
        Sort items by the given fields recursively in the order given.

        :param items_grouped: Map of items grouped by the last sort value.
        :param fields: Map of ``{<tag/property>: <reversed>}``.
            If reversed is True, sort the ``tag/property`` in reverse.
        :return: Map of grouped and sorted items.
        """
        field, reverse = next(iter(fields.items()), (None, None))
        if field is None:  # sorting complete
            return items_grouped

        fields = copy(fields)
        fields.pop(field)

        # sort each group and recurse through each field for each group
        for i, (key, items) in enumerate(items_grouped.items(), 1):
            cls.sort_by_field(items=items, field=field, reverse=reverse)
            items_grouped[key] = cls._sort_by_fields(cls.group_by_field(items, field=field), fields=fields)

        return items_grouped

    def as_dict(self):
        fields = None
        if isinstance(self.sort_fields, Mapping):
            fields = {field.name: "desc" if r else "asc" for field, r in self.sort_fields.items()}

        return {
            "sort_fields": fields,
            "shuffle_mode": self.shuffle_mode,
            "shuffle_weight": self.shuffle_weight
        }



"""
Processor operations that help a user to check whether the currently matched ID is valid for given items.

Provides the user the ability to modify associated IDs using a Remote player as an interface for
reviewing matches through temporary playlist creation.
"""
import itertools
import logging
from collections import Counter
from collections.abc import Sequence, Collection, Iterator, Awaitable
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
from typing import Any, Self

from musify import PROGRAM_NAME
from musify.base import MusifyItemSettable, Result
from musify.field import Fields
from musify.libraries.core.collection import MusifyCollection
from musify.libraries.remote.core.api import RemoteAPI
from musify.libraries.remote.core.factory import RemoteObjectFactory
from musify.libraries.remote.core.object import RemotePlaylist
from musify.libraries.remote.core.types import RemoteIDType, RemoteObjectType
from musify.logger import MusifyLogger
from musify.logger import REPORT
from musify.processors.base import InputProcessor
from musify.processors.match import ItemMatcher
from musify.processors.search import RemoteItemSearcher
from musify.utils import get_max_width, align_string

try:
    import tqdm
except ImportError:
    tqdm = None

ALLOW_KARAOKE_DEFAULT = RemoteItemSearcher.search_settings[RemoteObjectType.TRACK].allow_karaoke


@dataclass(frozen=True)
class ItemCheckResult[T: MusifyItemSettable](Result):
    """Stores the results of the checking process."""
    #: Sequence of Items that had URIs switched during the check.
    switched: Sequence[T] = field(default=tuple())
    #: Sequence of Items that were marked as unavailable.
    unavailable: Sequence[T] = field(default=tuple())
    #: Sequence of Items that were skipped from the check.
    skipped: Sequence[T] = field(default=tuple())


class RemoteItemChecker(InputProcessor):
    """
    Runs operations for checking the URIs associated with a collection of items.

    When running :py:func:`check`, the object will do the following:
        * Make temporary playlists for each collection up to a ``interval`` limit of playlists.
          At which point, playlist creation pauses.
        * User can then check and modify the temporary playlists to match items to correct items or otherwise.
        * When operations resume at the user's behest, the program will search each playlist to find changes
          and attempt to match any new items to a source item.
        * If no matches are found for certain items, the program will prompt the user
          to determine how they wish to deal with these items.
        * Operation completes once user exists or all items have an associated URI.

    :param matcher: The :py:class:`ItemMatcher` to use when comparing any changes made by the user in remote playlists
        during the checking operation
    :param object_factory: The :py:class:`RemoteObjectFactory` to use when creating new remote objects.
        This must have a :py:class:`RemoteAPI` assigned for this processor to work as expected.
    :param interval: Stop creating playlists after this many playlists have been created and pause for user input.
    :param allow_karaoke: When True, items determined to be karaoke are allowed when matching switched items.
        Skip karaoke results otherwise. Karaoke items are identified using the ``karaoke_tags`` attribute.
    """

    __slots__ = (
        "logger",
        "matcher",
        "factory",
        "interval",
        "allow_karaoke",
        "_playlist_originals",
        "_playlist_check_collections",
        "_started",
        "_skip",
        "_quit",
        "_remaining",
        "_switched",
        "_final_switched",
        "_final_unavailable",
        "_final_skipped",
    )

    @property
    def api(self) -> RemoteAPI:
        """The :py:class:`RemoteAPI` to call"""
        return self.factory.api

    def __init__(
            self,
            matcher: ItemMatcher,
            object_factory: RemoteObjectFactory,
            interval: int = 10,
            allow_karaoke: bool = ALLOW_KARAOKE_DEFAULT
    ):
        super().__init__()

        # noinspection PyTypeChecker
        #: The :py:class:`MusifyLogger` for this  object
        self.logger: MusifyLogger = logging.getLogger(__name__)

        #: The :py:class:`ItemMatcher` to use when comparing any changes made by the user in remote playlists
        #: during the checking operation
        self.matcher = matcher
        #: The :py:class:`RemoteObjectFactory` to use when creating new remote objects.
        self.factory = object_factory
        #: Stop creating playlists after this many playlists have been created and pause for user input
        self.interval = interval
        #: Allow karaoke items when matching on switched items
        self.allow_karaoke = allow_karaoke

        #: Map of playlist names to their RemotePlaylist object for check playlists.
        #: This stores the RemotePlaylist's state as it was when it was found
        #: if it already existed when called by the processor
        self._playlist_originals: dict[str, RemotePlaylist] = {}
        #: Map of playlist names to the collection of items added for check playlists
        self._playlist_check_collections: dict[str, MusifyCollection] = {}

        #: Whether a check was started
        self._started = False
        #: When true, skip the current loop and eventually safely quit check
        self._skip = False
        #: When true, safely quit check
        self._quit = False

        #: The currently remaining items that the user needs to manually check
        self._remaining: list[MusifyItemSettable] = []
        #: The list of items with switched URIs currently processed by the checker
        self._switched: list[MusifyItemSettable] = []

        #: The final list of items with switched URIs as processed by the checker
        self._final_switched: list[MusifyItemSettable] = []
        #: The final list of items marked as unavailable for this remote source
        self._final_unavailable: list[MusifyItemSettable] = []
        #: The final list of items skipped by the checker
        self._final_skipped: list[MusifyItemSettable] = []

    async def __aenter__(self) -> Self:
        await self.api.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.api.__aexit__(exc_type, exc_val, exc_tb)

    async def _check_api(self) -> None:
        """Check if the API token has expired and refresh as necessary"""
        await self.api.authorise()

    async def _create_playlist(self, collection: MusifyCollection[MusifyItemSettable]) -> None:
        """Create a temporary playlist, store its URL for later unfollowing, and add all given URIs."""
        await self._check_api()

        items = [item for item in collection if item.has_uri]
        if not items:
            return

        response = await self.api.get_or_create_playlist(collection.name, public=False)
        await self.api.follow_playlist(response)
        await self.api.extend_items(response=response, kind=RemoteObjectType.PLAYLIST, key=RemoteObjectType.TRACK)
        playlist: RemotePlaylist = self.factory.playlist(response=response)

        self._playlist_originals[collection.name] = playlist
        self._playlist_check_collections[collection.name] = collection

        await playlist.sync(items=items, kind="new", reload=False, dry_run=False)

    async def _delete_playlists(self) -> None:
        """Delete all temporary playlists stored and clear stored playlists and collections"""
        # assume all empty original playlists were temp playlists and delete them, restore the others
        delete_count = sum(1 for pl in self._playlist_originals.values() if len(pl) == 0)
        restore_count = sum(1 for pl in self._playlist_originals.values() if len(pl) > 0)
        if not delete_count + restore_count:
            return

        self.logger.info_extra(
            f"\33[93mDeleting {delete_count} temporary playlists and restoring {restore_count} playlists... \33[0m"
        )

        async def _process_playlist(pl: RemotePlaylist) -> None:
            if len(pl) == 0:
                await self.api.delete_playlist(pl)
            else:
                await pl.sync(kind="sync", reload=False, dry_run=False)

        await self._check_api()
        await self.logger.get_asynchronous_iterator(
            map(_process_playlist, self._playlist_originals.values()), disable=True
        )

        self._playlist_originals.clear()
        self._playlist_check_collections.clear()

    def __call__[T: MusifyItemSettable](
            self, collections: Collection[MusifyCollection[T]]
    ) -> Awaitable[ItemCheckResult[T] | None]:
        return self.check(collections)

    async def check[T: MusifyItemSettable](
            self, collections: Collection[MusifyCollection[T]]
    ) -> ItemCheckResult[T] | None:
        """
        Run the checker for the given ``collections``.

        :param collections: A list of collections to check.
        :return: A :py:class:`ItemCheckResult` object containing the remapped items created during the check.
            Return None when the user opted to quit (not skip) the checker before completion.
        """
        if len(collections) == 0 or all(not collection.items for collection in collections):
            self.logger.debug("\33[93mNo items to check. \33[0m")
            return

        self.logger.debug("Checking items: START")
        self.logger.info(
            f"\33[1;95m ->\33[1;97m Checking items by creating temporary {self.api.source} playlists "
            f"for the current user: {self.api.user_name} \33[0m"
        )

        total = len(collections)
        pages_total = (total // self.interval) + (total % self.interval > 0)
        bar = self.logger.get_synchronous_iterator(
            iter(collections), total=len(collections), desc="Creating temp playlists", unit="playlists"
        )

        self._started = True
        self._skip = False
        self._quit = False

        for page in range(1, pages_total + 1):
            try:
                await self.logger.get_asynchronous_iterator(
                    map(self._create_playlist, itertools.islice(bar, self.interval)), disable=True,
                )
                await self._pause(page=page, total=pages_total)

                if not self._quit:  # still run if skip is True
                    await self._check_uri()

            except KeyboardInterrupt:
                self.logger.error("User triggered exit with KeyboardInterrupt")
                self._quit = True
            finally:
                await self._delete_playlists()

            if self._quit or self._skip:  # quit check
                break

        result = await self.close()
        self.logger.debug("Checking items: DONE\n")
        return result

    async def close(self) -> ItemCheckResult | None:
        """Close the checker, deleting/syncing all active playlists and returning the result of the check."""
        await self._delete_playlists()
        if self._quit or not self._started:
            self._reset()
            return

        self.logger.print_line()
        self.logger.report(
            f"\33[1;96mCHECK TOTALS \33[0m| "
            f"\33[94m{len(self._final_switched):>5} switched  \33[0m| "
            f"\33[91m{len(self._final_unavailable):>5} unavailable \33[0m| "
            f"\33[93m{len(self._final_skipped):>5} skipped \33[0m"
        )
        self.logger.print_line(REPORT)

        result = ItemCheckResult(
            switched=self._final_switched, unavailable=self._final_unavailable, skipped=self._final_skipped
        )

        self._reset()
        return result

    def _reset(self):
        self._started = False
        self._remaining.clear()
        self._switched.clear()
        self._final_switched = []
        self._final_unavailable = []
        self._final_skipped = []

    ###########################################################################
    ## Pause to check items in current temp playlists
    ###########################################################################
    async def _pause(self, page: int, total: int) -> None:
        """
        Initial pause after the ``interval`` limit of playlists have been created.

        :param page: The current page number i.e. the current number of pauses that have happened so far.
        :param total: The total number of pages (pauses) that will occur in this run.
        """
        header = [
            f"\t\33[1;94mTemporary playlists created on {self.api.source}.",
            f"You may now check the songs in each playlist on {self.api.source}. \33[0m"
        ]
        options = {
            "<Name of playlist>":
                "Print position, item name, URI, and URL from given link of items as originally added to temp playlist",
            f"<{self.api.source} URL/URI>":
                "Print position, item name, URI, and URL from given link (useful to check current status of playlist)",
            "<Return/Enter>":
                "Once you have checked all playlist's items, continue on and check for any switches by the user",
            "l": "List the names of the temporary playlists created",
            "s": "Check for changes on current playlists, but skip any remaining checks",
            "q": "Delete current temporary playlists and quit check",
            "h": "Show this dialogue again",
        }

        current_input = "START"
        help_text = self._format_help_text(options=options, header=header)

        self.logger.print_message("\n" + help_text)
        while current_input != '':  # while user has not hit return only
            current_input = self._get_user_input(f"Enter ({page}/{total})")
            pl_name = next((
                name for name in self._playlist_check_collections
                if current_input and current_input.casefold() in name.casefold()
            ), None)

            if current_input.casefold() == "h":  # print help text
                self.logger.print_message("\n" + help_text)

            elif current_input.casefold() == 's' or current_input.casefold() == 'q':  # quit/skip
                self._quit = current_input.casefold() == 'q' or self._quit
                self._skip = current_input.casefold() == 's' or self._skip
                break

            elif current_input.casefold() == 'l':
                for name in self._playlist_check_collections:
                    self.logger.print_message(f"\33[97m- \33[91m{name}\33[0m")

            elif pl_name:  # print originally added items
                items = [item for item in self._playlist_check_collections[pl_name] if item.has_uri]
                max_width = get_max_width(items)

                self.logger.print_message(f"\n\t\33[96mShowing items originally added to \33[94m{pl_name}\33[0m:\n")
                for i, item in enumerate(items, 1):
                    length = getattr(item, "length", 0)
                    self.api.print_item(
                        i=i, name=item.name, uri=item.uri, length=length, total=len(items), max_width=max_width
                    )
                self.logger.print_message()

            elif self.api.wrangler.validate_id_type(current_input):  # print URL/URI/ID result
                await self._check_api()
                await self.api.print_collection(current_input)

            elif current_input != "":
                self.logger.warning("Input not recognised.")

    ###########################################################################
    ## Match items user has added or removed
    ###########################################################################
    async def _check_uri(self) -> None:
        """Run operations to check that URIs are assigned to all the items in the current list of collections."""
        skip_hold = self._skip
        self._skip = False
        for name, collection in self._playlist_check_collections.items():
            self.matcher.log([name, f"{len(collection):>6} total items"], pad='>')

            while True:
                await self._match_to_remote(name=name)
                self._match_to_input(name=name)
                if not self._remaining:
                    break

            unavailable = tuple(item for item in collection if item.has_uri is False)
            skipped = tuple(item for item in collection if item.has_uri is None)

            self.matcher.log([name, f"{len(self._switched):>6} items switched"], pad='<')
            self.matcher.log([name, f"{len(unavailable):>6} items unavailable"])
            self.matcher.log([name, f"{len(skipped):>6} items skipped"])

            self._final_switched += self._switched
            self._final_unavailable += unavailable
            self._final_skipped += skipped
            self._switched.clear()

            if self._quit or self._skip:
                break

        self._skip = skip_hold

    async def _match_to_remote(self, name: str) -> None:
        """
        Check the current temporary playlist given by ``name`` and attempt to match the source list of items
        to any modifications the user has made.
        """
        await self._check_api()

        self.logger.info(
            "\33[1;95m ->\33[1;97m Checking for changes to items in "
            f"{self.api.source} playlist: \33[94m{name}\33[0m..."
        )

        source = self._playlist_check_collections[name]
        source_valid = [item for item in source if item.has_uri]

        pl_original = self._playlist_originals[name]
        remote_response = next(iter(await self.api.get_items(pl_original.url, extend=True)))
        remote = self.factory.playlist(response=remote_response)
        remote_valid = [item for item in remote if item.has_uri]

        added = [item for item in remote_valid if item not in source and item not in pl_original]
        removed = [
            item for item in source_valid if item not in remote_valid and item not in pl_original
        ] if not self._remaining else []
        missing = self._remaining or [item for item in source if item.has_uri is None]

        if len(added) + len(removed) + len(missing) == 0:
            if len(source_valid) == len(remote_valid):
                self.matcher.log([name, "Playlist unchanged and no missing URIs, skipping match"])
                return

            # if item collection originally contained duplicate URIS and one or more of the duplicates were removed,
            # find missing items by looking for changes in counts
            remote_counts = Counter(item.uri for item in remote_valid)
            for uri, count in Counter(item.uri for item in source_valid).items():
                if remote_counts.get(uri) != count:
                    missing.extend([item for item in source_valid if item.uri == uri])

        discount = sum(1 for item in remote if item in pl_original and item in source)
        self.matcher.log([name, f"{len(added):>6} items added"])
        self.matcher.log([name, f"{len(removed):>6} items removed"])
        self.matcher.log([name, f"{len(missing):>6} items in source missing URI"])
        self.matcher.log([name, f"{len(pl_original):>6} items in the playlist at start"])
        self.matcher.log([name, f"{discount:>6} discounted items from the source that were in the playlist at start"])
        self.matcher.log([name, f"{len(added) - len(removed):>6} total item changes"])

        remaining = removed + missing
        count_start = len(remaining)
        with ThreadPoolExecutor(thread_name_prefix="checker") as executor:
            tasks: Iterator[tuple[MusifyItemSettable, MusifyItemSettable | None]] = executor.map(
                lambda item: (
                    item, self.matcher(item, results=added, match_on=[Fields.TITLE], allow_karaoke=self.allow_karaoke)
                ),
                remaining if added else ()
            )

        for item, match in tasks:
            if not match:
                continue

            item.uri = match.uri

            if match in added:
                added.remove(match)
            removed.remove(item) if item in removed else missing.remove(item)
            self._switched.append(match)

        self._remaining = removed + missing
        count_final = len(self._remaining)
        self.matcher.log([name, f"{count_start - count_final:>6} items switched"])
        self.matcher.log([name, f"{count_final:>6} items still not found"])

    def _match_to_input(self, name: str) -> None:
        """
        Get the user's input for any items in the ``remaining`` list that are still missing URIs.
        Provide a ``name`` to use for logging only.
        """
        if not self._remaining:
            return

        header = [f"\t\33[1;94m{name}:\33[91m The following items were removed and/or matches were not found. \33[0m"]
        options = {
            f"<{self.api.source} ID/URL/URI>": "Assign the given ID/URL/URI to the item",
            "u": f"Mark item as 'Unavailable on {self.api.source}'",
            "n": f"Leave item with no URI. ({PROGRAM_NAME} will still attempt to find this item at the next run)",
            "a": "Add in addition to 'u' or 'n' options to apply this setting to all items in this playlist",
            "r": "Recheck playlist for all items in the collection",
            "p": "Print the local path of the current item if available",
            "s": "Skip checking process for all current playlists",
            "q": "Skip checking process for all current playlists and quit check",
            "h": "Show this dialogue again",
        }

        current_input = ""
        help_text = self._format_help_text(options=options, header=header)
        help_text += "OR enter a custom URI/URL/ID for this item\n"

        self.matcher.log([name, f"Getting user input for {len(self._remaining)} items"])
        max_width = get_max_width({item.name for item in self._remaining})

        self.logger.print_message("\n" + help_text)
        for item in self._remaining.copy():
            while current_input is not None and item in self._remaining:  # while item not matched or skipped
                self.matcher.log([name, f"{len(self._remaining):>6} remaining items"])
                if 'a' not in current_input:
                    current_input = self._get_user_input(align_string(item.name, max_width=max_width))

                if current_input.casefold() == 'h':  # print help
                    self.logger.print_message("\n" + help_text)
                else:
                    current_input = self._match_item_to_input(name=name, item=item, current_input=current_input)

            if current_input is None or not self._remaining:
                break

    def _match_item_to_input(self, name: str, item: MusifyItemSettable, current_input: str) -> str | None:
        if current_input.casefold() == 's' or current_input.casefold() == 'q':  # quit/skip
            self.matcher.log([name, "Skipping all loops"], pad="<")
            self._quit = current_input.casefold() == 'q' or self._quit
            self._skip = current_input.casefold() == 's' or self._skip
            self._remaining.clear()
            return

        elif current_input.casefold().replace('a', '') == 'u':  # mark item as unavailable
            self.matcher.log([name, "Marking as unavailable"], pad="<")
            item.uri = self.api.wrangler.unavailable_uri_dummy
            self._remaining.remove(item)

        elif current_input.casefold().replace('a', '') == 'n':  # leave item without URI and unprocessed
            self.matcher.log([name, "Skipping"], pad="<")
            item.uri = None
            self._remaining.remove(item)

        elif current_input.casefold() == 'r':  # return to former 'while' loop
            self.matcher.log([name, "Refreshing playlist metadata and restarting loop"])
            return

        elif current_input.casefold() == 'p' and hasattr(item, "path"):  # print item path
            self.logger.print_message(f"\33[96m{item.path}\33[0m")

        elif self.api.wrangler.validate_id_type(current_input):  # update URI and add item to switched list
            uri = self.api.wrangler.convert(
                current_input, kind=RemoteObjectType.TRACK, type_out=RemoteIDType.URI
            )

            self.matcher.log([name, f"Updating URI: {item.uri} -> {uri}"], pad="<")
            item.uri = uri

            self._switched.append(item)
            self._remaining.remove(item)
            current_input = ""

        elif current_input:  # invalid input
            self.logger.warning("Input not recognised.")
            current_input = ""

        return current_input

    def as_dict(self) -> dict[str, Any]:
        return {
            "matcher": self.matcher,
            "remote_source": self.factory.api.source,
            "interval": self.interval,
            "allow_karaoke": self.allow_karaoke,
        }



"""
All core type hints to use throughout the entire package.
"""
from collections.abc import Iterable, Sequence, MutableSequence, Collection
from enum import IntEnum
from typing import Self, Any

from musify.exception import MusifyEnumError

type UnitIterable[T] = T | Iterable[T]
type UnitCollection[T] = T | Collection[T]
type UnitSequence[T] = T | Sequence[T]
type UnitMutableSequence[T] = T | MutableSequence[T]
type UnitList[T] = T | list[T]

Number = int | float


class MusifyEnum(IntEnum):
    """Generic class for :py:class:`IntEnum` implementations for the entire package."""

    @staticmethod
    def _unique_list(value: Iterable[Any]) -> list[Any]:
        """
        Returns a copy of the given ``value`` that contains only unique elements.
        Useful for producing unique lists whilst preserving order.
        """
        unique = []
        for item in value:
            if item not in unique:
                unique.append(item)
        return unique

    @classmethod
    def map(cls, enum: Self) -> list[Self]:
        """
        Optional mapper to apply to the enum found during :py:meth:`all`, :py:meth:`from_name`,
        and :py:meth:`from_value` calls
        """
        return [enum]

    @classmethod
    def all(cls) -> list[Self]:
        """Get all enums for this enum."""
        return cls._unique_list(e for enum in cls if enum.name != "ALL" for e in cls.map(enum))

    @classmethod
    def from_name(cls, *names: str, fail_on_many: bool = True) -> list[Self]:
        """
        Returns all enums that match the given enum names

        :param fail_on_many: If more than one enum is found, raise an exception.
        :raise EnumNotFoundError: If a corresponding enum cannot be found.
        """
        names_upper = [name.strip().upper() for name in names]
        enums = cls._unique_list(e for enum in cls if enum.name in names_upper for e in cls.map(enum))

        if len(enums) == 0:
            raise MusifyEnumError(names)
        elif len(enums) > 1 and fail_on_many:
            raise MusifyEnumError(value=enums, message="Too many enums found")

        return enums

    @classmethod
    def from_value(cls, *values: int, fail_on_many: bool = True) -> list[Self]:
        """
        Returns all enums that match the given enum values

        :param fail_on_many: If more than one enum is found, raise an exception.
        :raise EnumNotFoundError: If a corresponding enum cannot be found.
        """
        enums = cls._unique_list(e for enum in cls if enum.value in values for e in cls.map(enum))
        if len(enums) == 0:
            raise MusifyEnumError(values)
        elif len(enums) > 1 and fail_on_many:
            raise MusifyEnumError(value=enums, message="Too many enums found")
        return enums



"""
Processor that converts representations of time units to python time objects.
"""
from datetime import timedelta
from typing import Any

from dateutil.relativedelta import relativedelta

from musify.printer import PrettyPrinter
from musify.processors.base import DynamicProcessor, dynamicprocessormethod


class TimeMapper(DynamicProcessor, PrettyPrinter):
    """Map of time character representation to it unit conversion from seconds"""

    __slots__ = ()

    @classmethod
    def _processor_method_fmt(cls, name: str) -> str:
        return name.casefold().strip()[0] if not name.startswith("min") else name

    def __init__(self, func: str):
        super().__init__()
        self._set_processor_name(func)

    def __call__(self, *args, **kwargs) -> timedelta | relativedelta:
        return self.map(*args, **kwargs)

    def map(self, value: Any):
        """Run the mapping function"""
        return super().__call__(value)

    @dynamicprocessormethod
    def seconds(self, value: Any) -> timedelta:
        """Map given ``value`` in seconds to :py:class:`timedelta`"""
        return timedelta(seconds=int(value))

    @dynamicprocessormethod("min")
    def minutes(self, value: Any) -> timedelta:
        """Map given ``value`` in minutes to :py:class:`timedelta`"""
        return timedelta(minutes=int(value))

    @dynamicprocessormethod
    def hours(self, value: Any) -> timedelta:
        """Map given ``value`` in hours to :py:class:`timedelta`"""
        return timedelta(hours=int(value))

    @dynamicprocessormethod
    def days(self, value: Any) -> timedelta:
        """Map given ``value`` in days to :py:class:`timedelta`"""
        return timedelta(days=int(value))

    @dynamicprocessormethod
    def weeks(self, value: Any) -> timedelta:
        """Map given ``value`` in weeks to :py:class:`timedelta`"""
        return timedelta(weeks=int(value))

    @dynamicprocessormethod
    def months(self, value: Any) -> relativedelta:
        """Map given ``value`` in months to :py:class:`timedelta`"""
        return relativedelta(months=int(value))

    def as_dict(self) -> dict[str, Any]:
        return {"function": self._processor_name}



"""
The core, basic library implementation which is just a simple set of folders.
"""
import functools
import itertools
import os
from collections.abc import Collection, Mapping, Iterable
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Any

from musify.base import Result
from musify.exception import MusifyError
from musify.file.path_mapper import PathMapper, PathStemMapper
from musify.libraries.core.object import Library, LibraryMergeType
from musify.libraries.local.collection import LocalCollection, LocalFolder, LocalAlbum, LocalArtist, LocalGenres
from musify.libraries.local.playlist import PLAYLIST_CLASSES, LocalPlaylist, load_playlist
from musify.libraries.local.track import TRACK_CLASSES, LocalTrack, load_track
from musify.libraries.local.track.field import LocalTrackField
from musify.libraries.remote.core.wrangle import RemoteDataWrangler
from musify.logger import STAT
from musify.processors.base import Filter
from musify.processors.filter import FilterDefinedList
from musify.processors.sort import ItemSorter
from musify.types import UnitCollection, UnitIterable
from musify.utils import align_string, get_max_width, to_collection

type RestoreTracksType = Iterable[Mapping[str, Any]] | Mapping[str | Path, Mapping[str, Any]]


class LocalLibrary(LocalCollection[LocalTrack], Library[LocalTrack]):
    """
    Represents a local library, providing various methods for manipulating
    tracks and playlists across an entire local library collection.

    :param library_folders: The absolute paths of the library folders containing all tracks.
        The intialiser will check for the existence of these paths and only store them if they exist.
    :param playlist_folder: The absolute path of the playlist folder containing all playlists
        or the relative path within one of the available ``library_folders``.
        If a relative path is given and many library folders are given,
        only the first path that gives an existing result is processed.
        The setter will check for the existence of this path and only store the absolute path if it exists.
    :param playlist_filter: An optional :py:class:`Filter` to apply or collection of playlist names to include when
        loading playlists. Playlist names will be passed to this filter to limit which playlists are loaded.
    :param path_mapper: Optionally, provide a :py:class:`PathMapper` for paths stored in the playlist files.
        Useful if the playlist files contain relative paths and/or paths for other systems that need to be
        mapped to absolute, system-specific paths to be loaded and back again when saved.
    :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object for processing URIs on tracks.
        If given, the wrangler can be used when calling __get_item__ to get an item from the collection from its URI.
        The wrangler is also used when loading tracks to allow them to process URI tags.
        For more info on this, see :py:class:`LocalTrack`.
    :param name: A name to assign to this library.
    """

    __slots__ = (
        "_name",
        "_library_folders",
        "_playlist_folder",
        "playlist_filter",
        "path_mapper",
        "_playlist_paths",
        "_playlists",
        "_track_paths",
        "_tracks",
        "errors",
    )
    __attributes_classes__ = (Library, LocalCollection)
    __attributes_ignore__ = ("tracks_in_playlists",)

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    # noinspection PyPropertyDefinition
    @classmethod
    @property
    def source(cls) -> str:
        """The type of local library loaded"""
        return cls.__name__.replace("Library", "")

    @property
    def tracks(self) -> list[LocalTrack]:
        """The tracks in this collection"""
        return self._tracks

    @property
    def playlists(self) -> dict[str, LocalPlaylist]:
        """The playlists in this library mapped as ``{<name>: <playlist>}``"""
        return self._playlists

    @property
    def library_folders(self) -> list[Path]:
        """Path to the library folder"""
        return self._library_folders

    @library_folders.setter
    def library_folders(self, folders: UnitCollection[str | Path] | None):
        """
        Sets the library folder path and generates a set of available and valid track paths in the folder.
        Skips settings if the given value is None.
        """
        if folders is None:
            return

        self._library_folders: list[Path] = [folder for folder in map(Path, to_collection(folders)) if folder.exists()]

        self._track_paths = {
            path for folder in self._library_folders for cls in TRACK_CLASSES for path in cls.get_filepaths(folder)
        }
        if isinstance(self.path_mapper, PathStemMapper):
            self.path_mapper.available_paths = self._track_paths

        self.logger.debug(
            f"Set library folder(s): {", ".join(map(str, self.library_folders))} | "
            f"{len(self._track_paths)} track paths found"
        )

    @property
    def playlist_folder(self) -> Path:
        """Path to the playlist folder"""
        return self._playlist_folder

    @playlist_folder.setter
    def playlist_folder(self, folder: str | Path | None):
        """
        Sets the playlist folder path and generates a set of available and valid playlists in the folder.
        Appends the library folder path if the given path is not valid. Skips settings if the given value is None.
        """
        if folder is None:
            return

        folder = Path(folder)
        if (not folder.is_dir() or not folder.is_absolute()) and self.library_folders:
            for fldr in self.library_folders:
                value = fldr.joinpath(folder)
                if value.is_dir():
                    folder = value
                    break
        if not folder.is_dir():
            return

        self._playlist_folder = folder
        self._playlist_paths = None

        playlists = {
            Path(str(path).removeprefix(str(self._playlist_folder))).stem: path
            for cls in PLAYLIST_CLASSES for path in cls.get_filepaths(self._playlist_folder)
        }

        pl_total = len(playlists)
        pl_filtered = self.playlist_filter(playlists)
        self._playlist_paths = {
            name: path for name, path in sorted(playlists.items(), key=lambda x: x[0].casefold())
            if name in pl_filtered
        }

        self.logger.debug(f"Set playlist folder: {self.playlist_folder} | " + (
            f"Filtered out {pl_total - len(self._playlist_paths)} playlists "
            f"from {pl_total} {self.name} available playlists"
            if (pl_total - len(pl_filtered)) > 0 else f"{len(self._playlist_paths)} playlists found"
        ))

    @property
    def folders(self) -> list[LocalFolder]:
        """
        Dynamically generate a set of folder collections from the tracks in this library.
        Folder collections are generated relevant to the library folder it is found in.
        """
        def get_relative_path(track: LocalTrack) -> Path:
            """Return path of a track relative to the library folders of this library"""
            path = functools.reduce(
                lambda p, f: str(p).replace(str(f), ""), self.library_folders, str(track.path)
            )
            return Path(path.lstrip(os.path.sep)).parent

        def create_folder_collection(path: Path, tracks: Collection[LocalTrack]) -> LocalFolder:
            """
            Create a :py:class:`LocalFolder` collection from the given ``tracks``,
            ensuring the collection has the exact given ``name``.

            The LocalFolder filters the input tracks by the given name,
            where it will match on ``track.folder`` == ``name``.
            Given that ``track.folder`` is just the direct parent folder stem only,
            We need to supply this stem folder on init and assign the full folder path back after.
            """
            folder = LocalFolder(tracks=tracks, name=path.name, remote_wrangler=self.remote_wrangler)
            folder._name = str(path)
            return folder

        grouped = itertools.groupby(sorted(self.tracks, key=lambda track: track.path), get_relative_path)
        collections = [create_folder_collection(path=path, tracks=list(group)) for path, group in grouped]
        return sorted(collections, key=lambda x: x.name)

    @property
    def albums(self) -> list[LocalAlbum]:
        """Dynamically generate a set of album collections from the tracks in this library"""
        grouped = ItemSorter.group_by_field(items=self.tracks, field=LocalTrackField.ALBUM)
        collections = [
            LocalAlbum(tracks=group, name=name, remote_wrangler=self.remote_wrangler)
            for name, group in grouped.items() if name
        ]
        return sorted(collections, key=lambda x: x.name)

    @property
    def artists(self) -> list[LocalArtist]:
        """Dynamically generate a set of artist collections from the tracks in this library"""
        grouped = ItemSorter.group_by_field(items=self.tracks, field=LocalTrackField.ARTISTS)
        collections = [
            LocalArtist(tracks=group, name=name, remote_wrangler=self.remote_wrangler)
            for name, group in grouped.items() if name
        ]
        return sorted(collections, key=lambda x: x.name)

    @property
    def genres(self) -> list[LocalGenres]:
        """Dynamically generate a set of genre collections from the tracks in this library"""
        grouped = ItemSorter.group_by_field(items=self.tracks, field=LocalTrackField.GENRES)
        collections = [
            LocalGenres(tracks=group, name=name, remote_wrangler=self.remote_wrangler)
            for name, group in grouped.items() if name
        ]
        return sorted(collections, key=lambda x: x.name)

    def __init__(
            self,
            library_folders: UnitCollection[str | Path] | None = None,
            playlist_folder: str | Path | None = None,
            playlist_filter: Collection[str] | Filter[str] = (),
            path_mapper: PathMapper = PathMapper(),
            remote_wrangler: RemoteDataWrangler | None = None,
            name: str = None,
    ):
        super().__init__(remote_wrangler=remote_wrangler)

        self._name: str = name if name else self.source

        self.logger.debug(f"Setup {self.name} library: START")
        self.logger.info(f"\33[1;95m ->\33[1;97m Setting up {self.name} library \33[0m")
        self.logger.print_line()

        #: Passed to playlist objects when loading playlists to map paths stored in the playlist file.
        self.path_mapper = path_mapper

        self._library_folders: list[Path] = []
        self._track_paths: set[Path] = set()
        self.library_folders = library_folders

        if not isinstance(playlist_filter, Filter):
            playlist_filter = FilterDefinedList(playlist_filter)
        #: :py:class:`Filter` to filter out the playlists loaded by name.
        self.playlist_filter: Filter[str] = playlist_filter

        self._playlist_folder: Path | None = None
        # playlist lowercase name mapped to its filepath for all accepted filetypes in playlist folder
        self._playlist_paths: dict[str, Path] = {}
        self.playlist_folder = playlist_folder

        self._tracks: list[LocalTrack] = []
        self._playlists: dict[str, LocalPlaylist] = {}

        #: Stores the paths that caused errors when loading/enriching
        self.errors: list[str] = []
        self.logger.debug(f"Setup {self.name} library: DONE\n")

    async def load(self) -> None:
        """Loads all tracks and playlists in this library from scratch and log results."""
        self.logger.debug(f"Load {self.name} library: START")
        self.logger.info(
            f"\33[1;95m ->\33[1;97m Loading {self.name} library of "
            f"{len(self._track_paths)} tracks and {len(self._playlist_paths)} playlists \33[0m"
        )

        await self.load_tracks()
        await self.load_playlists()

        self.logger.print_line(STAT)
        self.log_tracks()
        self.log_playlists()

        self.logger.print_line()
        self.logger.debug(f"Load {self.name} library: DONE\n")

    def _log_errors(self, message: str = "Could not load") -> None:
        """Log paths which had some error while loading"""
        errors = tuple(f"\33[91m{e}\33[0m" for e in sorted(self.errors))
        if len(errors) > 0:
            self.logger.warning(f"\33[97m{message}: \33[0m\n\t- {"\n\t- ".join(errors)} ")
            self.logger.print_line()
        self.errors.clear()

    ###########################################################################
    ## Tracks
    ###########################################################################
    async def load_track(self, path: str | Path) -> LocalTrack | None:
        """
        Wrapper for :py:func:`load_track` which automatically loads the track at the given ``path``
        and assigns optional arguments using this library's attributes.

        Handles exceptions by logging paths which produce errors to internal list of ``errors``.
        """
        try:
            return await load_track(path=path, remote_wrangler=self.remote_wrangler)
        except MusifyError as ex:
            self.logger.debug(f"Load error for track: {path} - {ex}")
            self.errors.append(path)

    async def load_tracks(self) -> None:
        """Load all tracks from all the valid paths in this library, replacing currently loaded tracks."""
        if not self._track_paths:
            return

        self.logger.debug(f"Load {self.name} tracks: START")
        self.logger.info(
            f"\33[1;95m  >\33[1;97m Extracting metadata and properties for {len(self._track_paths)} tracks \33[0m"
        )

        # WARNING: making this run asynchronously will break tqdm; bar will get stuck after 1-2 ticks
        bar = self.logger.get_synchronous_iterator(
            self._track_paths,
            desc="Loading tracks",
            unit="tracks",
            total=len(self._track_paths)
        )
        self._tracks = [await self.load_track(path) for path in bar]

        self._log_errors("Could not load the following tracks")
        self.logger.debug(f"Load {self.name} tracks: DONE\n")

    def log_tracks(self) -> None:
        width = get_max_width(self._playlist_paths) if self._playlist_paths else 20
        self.logger.stat(
            f"\33[1;96m{'LIBRARY URIS':<{width}}\33[1;0m |"
            f"\33[92m{sum([track.has_uri is True for track in self.tracks]):>6} available \33[0m|"
            f"\33[91m{sum([track.has_uri is None for track in self.tracks]):>6} missing \33[0m|"
            f"\33[93m{sum([track.has_uri is False for track in self.tracks]):>6} unavailable \33[0m|"
            f"\33[1;94m{len(self.tracks):>6} total \33[0m"
        )

    ###########################################################################
    ## Playlists
    ###########################################################################
    async def load_playlist(self, path: str | Path) -> LocalPlaylist:
        """
        Wrapper for :py:func:`load_playlist` which automatically loads the playlist at the given ``path``
        and assigns optional arguments using this library's attributes.

        Handles exceptions by logging paths which produce errors to internal list of ``errors``.
        """
        try:
            return await load_playlist(
                path=path, tracks=self.tracks, path_mapper=self.path_mapper, remote_wrangler=self.remote_wrangler,
            )
        except MusifyError as ex:
            self.logger.debug(f"Load error for playlist: {path} - {ex}")
            self.errors.append(path)

    async def load_playlists(self) -> None:
        """
        Load all playlists found in this library's ``playlist_folder``,
        filtered down using the ``playlist_filter`` if given, replacing currently loaded playlists.

        :return: The loaded playlists.
        :raise LocalCollectionError: If a given playlist name cannot be found.
        """
        if not self._playlist_paths:
            return

        self.logger.debug(f"Load {self.name} playlist data: START")
        self.logger.info(
            f"\33[1;95m  >\33[1;97m Loading playlist data for {len(self._playlist_paths)} playlists \33[0m"
        )

        # WARNING: making this run asynchronously will break tqdm; bar will get stuck after 1-2 ticks
        bar = self.logger.get_synchronous_iterator(
            self._playlist_paths.values(),
            desc="Loading playlists",
            unit="playlists",
            total=len(self._playlist_paths)
        )
        self._playlists = {
            pl.name: pl for pl in sorted([await self.load_playlist(pl) for pl in bar], key=lambda x: x.name.casefold())
        }

        self._log_errors("Could not load the following playlists")
        self.logger.debug(f"Load {self.name} playlists: DONE\n")

    def log_playlists(self) -> None:
        max_width = get_max_width(self.playlists)

        self.logger.stat(f"\33[1;96m{self.name.upper()} PLAYLISTS: \33[0m")
        for name, playlist in self.playlists.items():
            self.logger.stat(
                f"\33[97m{align_string(name, max_width=max_width)} \33[0m|"
                f"\33[92m{sum(1 for t in playlist if t.has_uri):>6} available \33[0m|"
                f"\33[91m{sum(1 for t in playlist if t.has_uri is None):>6} missing \33[0m|"
                f"\33[93m{sum(1 for t in playlist if t.has_uri is False):>6} unavailable \33[0m|"
                f"\33[1;94m{len(playlist):>6} total \33[0m"
            )

    async def save_playlists(self, dry_run: bool = True) -> dict[LocalPlaylist, Result]:
        """
        For each Playlist in this Library, saves its associate tracks and its settings (if applicable) to file.

        :param dry_run: Run function, but do not modify the file on the disk.
        :return: A map of the playlist name to the results of its sync as a :py:class:`Result` object.
        """
        async def _save_playlist(pl: LocalPlaylist) -> tuple[LocalPlaylist, Result]:
            return pl, await pl.save(dry_run=dry_run)

        # WARNING: making this run asynchronously will break tqdm; bar will get stuck after 1-2 ticks
        bar = self.logger.get_synchronous_iterator(
            self.playlists.values(), desc="Updating playlists", unit="tracks"
        )
        return dict([await _save_playlist(pl) for pl in bar])

    def merge_playlists(
            self, playlists: LibraryMergeType[LocalTrack], reference: LibraryMergeType[LocalTrack] | None = None
    ) -> None:
        current_names = set(self.playlists)

        super().merge_playlists(playlists=playlists, reference=reference)

        for pl in self.playlists.values():
            if pl.name in current_names:
                continue

            if isinstance(playlists, LocalLibrary):
                rel_path = pl.path.relative_to(playlists.playlist_folder)
            else:
                rel_path = pl.path.name

            pl.path = self.playlist_folder.joinpath(rel_path)

    ###########################################################################
    ## Backup/restore
    ###########################################################################
    def restore_tracks(
            self, backup: RestoreTracksType, tags: UnitIterable[LocalTrackField] = LocalTrackField.ALL
    ) -> int:
        """
        Restore track tags from a backup to loaded track objects. This does not save the updated tags.

        :param backup: Backup data in the form ``{<path>: {<Map of JSON formatted track data>}}``
        :param tags: Set of tags to restore.
        :return: The number of tracks restored
        """
        backup = self._extract_tracks_from_backup(backup)
        tag_names = set(LocalTrackField.to_tags(tags))

        count = 0
        for track in self.tracks:
            track_map = backup.get(track.path)
            if not track_map:
                continue

            for tag in tag_names:
                if tag in track_map:
                    track[tag] = track_map[tag]
            count += 1

        return count

    @staticmethod
    def _extract_tracks_from_backup(backup: RestoreTracksType) -> dict[Path, Mapping[str, Any]]:
        if isinstance(backup, Mapping):
            backup = {Path(path): track_map for path, track_map in backup.items()}
        else:
            backup = {Path(track_map["path"]): track_map for track_map in backup}
        return backup

    def _get_attributes(self) -> dict[str, Any]:
        attributes_extra = {"remote_source": self.remote_wrangler.source if self.remote_wrangler else None}
        return super()._get_attributes() | attributes_extra

    def json(self):
        self.logger.info(
            f"\33[1;95m ->\33[1;97m Extracting JSON data for library of "
            f"{len(self.tracks)} tracks and {len(self.playlists)} playlists\n"
        )

        attributes = self._get_attributes()

        playlists: dict[str, LocalPlaylist] | None = None
        if "playlists" in attributes and "tracks" in attributes:
            playlists = attributes["playlists"]
            attributes["playlists"] = {}

        self_json = self._to_json(attributes, pool=True)

        if playlists is not None:
            tracks: Mapping[str, Mapping[str, Any]] = {track["path"]: track for track in self_json["tracks"]}

            def _get_playlist_json(pl: LocalPlaylist) -> tuple[str, dict[str, Any]]:
                pl_attributes = pl._get_attributes()
                pl_attributes["tracks"] = []

                pl_json = pl._to_json(pl_attributes, pool=True)
                pl_json["tracks"] = [tracks.get(str(track.path), str(track.path)) for track in pl]

                return pl.name, pl_json

            with ThreadPoolExecutor(thread_name_prefix="to-json-playlists") as executor:
                tasks = executor.map(_get_playlist_json, playlists.values())

            self_json["playlists"] = dict(sorted(tasks, key=lambda x: x[0]))

        return self_json



"""
Implements all :py:mod:`Remote` object types for Spotify.
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections.abc import Iterable, MutableMapping, Mapping, Collection
from copy import copy, deepcopy
from datetime import datetime
from typing import Any, Self

from musify.libraries.remote.core import RemoteResponse
from musify.libraries.remote.core.object import RemoteCollectionLoader, RemoteTrack
from musify.libraries.remote.core.object import RemotePlaylist, RemoteAlbum, RemoteArtist
from musify.libraries.remote.core.types import APIInputValueSingle, RemoteIDType, RemoteObjectType
from musify.libraries.remote.spotify.api import SpotifyAPI
from musify.libraries.remote.spotify.base import SpotifyObject, SpotifyItem
from musify.libraries.remote.spotify.exception import SpotifyCollectionError
from musify.types import UnitCollection
from musify.utils import to_collection


class SpotifyTrack(SpotifyItem, RemoteTrack):
    """
    Extracts key ``track`` data from a Spotify API JSON response.

    :param response: The Spotify API JSON response.
    """

    __slots__ = ("_disc_total", "_comments", "_artists")

    _song_keys = ("C", "C#/Db", "D", "D#/Eb", "E", "F", "F#/Gb", "G", "G#/Ab", "A", "A#/Bb", "B")

    @property
    def name(self):
        return self.title

    @property
    def title(self) -> str:
        return self.response["name"]

    @property
    def artist(self):
        artists = self.response.get("artists", {})
        artist = self.tag_sep.join(artist["name"] for artist in artists)
        return artist or None

    @property
    def artists(self) -> list[SpotifyArtist]:
        return self._artists

    @property
    def album(self):
        return self.response.get("album", {}).get("name")

    @property
    def album_artist(self):
        album = self.response.get("album", {})
        album_artist = self.tag_sep.join(artist["name"] for artist in album.get("artists", []))
        return album_artist or None

    @property
    def track_number(self) -> int:
        return self.response["track_number"]

    @property
    def track_total(self):
        return self.response.get("album", {}).get("total_tracks")

    @property
    def genres(self):
        """
        List of genres for the album this track is featured on.
        If not found, genres from the main artist are given.
        """
        album = self.response.get("album", {})
        if album.get("genres"):
            return [g.title() for g in album.get("genres")]

        artists = self.response.get("artists", {})
        if not artists:
            return

        main_artist_genres = artists[0].get("genres", [])
        return [g.title() for g in main_artist_genres] if main_artist_genres else None

    @property
    def year(self):
        album = self.response.get("album", {})
        if "release_date" not in album:
            return
        values = album["release_date"].split("-")
        return int(values[0]) if len(values) >= 1 and values[0].isdigit() else None

    @property
    def month(self):
        album = self.response.get("album", {})
        if "release_date" not in album or "release_date_precision" not in album:
            return
        if album["release_date_precision"] in {"year"}:
            # release date is not precise to month, do not return month even if it exists in the value
            return
        values = album["release_date"].split("-")
        return int(values[1]) if len(values) >= 2 and values[1].isdigit() else None

    @property
    def day(self):
        album = self.response.get("album", {})
        if "release_date" not in album or "release_date_precision" not in album:
            return
        if album["release_date_precision"] in {"year", "month"}:
            # release date is not precise to day, do not return day even if it exists in the value
            return
        values = album["release_date"].split("-")
        return int(values[2]) if len(values) >= 3 and values[2].isdigit() else None

    @property
    def bpm(self):
        if "audio_features" not in self.response:
            return
        return self.response["audio_features"]["tempo"]

    @property
    def key(self):
        if "audio_features" not in self.response:
            return

        # correctly formatted song key string
        key_value: int = self.response["audio_features"]["key"]
        key: str | None = self._song_keys[key_value] if key_value >= 0 else None
        is_minor: bool = self.response["audio_features"]["mode"] == 0

        if not key:
            return None
        elif '/' in key:
            key_sep = key.split('/')
            return f"{key_sep[0]}{'m'*is_minor}/{key_sep[1]}{'m'*is_minor}"
        else:
            return f"{key}{'m'*is_minor}"

    @property
    def disc_number(self) -> int:
        return self.response["disc_number"]

    @property
    def disc_total(self):
        return self._disc_total

    @disc_total.setter
    def disc_total(self, value: int | None):
        self._disc_total = value

    @property
    def compilation(self) -> bool:
        album = self.response.get("album", {})
        return album.get("album_type", "") == "compilation"

    @property
    def comments(self):
        return self._comments

    @comments.setter
    def comments(self, value: UnitCollection[str] | None):
        self._comments = [value] if isinstance(value, str) else to_collection(value, list)

    @property
    def image_links(self):
        album = self.response.get("album", {})
        if not (images := album.get("images", [])):
            return {}

        images = {image["height"]: image["url"] for image in images}
        return {"cover_front": next(url for height, url in images.items() if height == max(images))}

    @property
    def length(self):
        if isinstance(self.response["duration_ms"], Mapping):
            return self.response["duration_ms"]["totalMilliseconds"] / 1000
        return self.response["duration_ms"] / 1000

    @property
    def rating(self):
        return self.response.get("popularity")

    def __init__(self, response: dict[str, Any], api: SpotifyAPI | None = None, skip_checks: bool = False):
        self._disc_total = None
        self._comments = None

        self._artists: list[SpotifyArtist] | None = None

        if "track" in response and isinstance(response["track"], dict):
            # happens in 'user's saved ...' or playlist responses
            response = response["track"]

        super().__init__(response=response, api=api, skip_checks=skip_checks)

    def refresh(self, skip_checks: bool = False) -> None:
        self._artists = [
            SpotifyArtist(artist, api=self.api, skip_checks=skip_checks)
            for artist in self._response.get("artists", {})
        ]

    @classmethod
    async def load(
            cls,
            value: APIInputValueSingle[Self],
            api: SpotifyAPI,
            features: bool = False,
            analysis: bool = False,
            extend_album: bool = False,
            extend_artists: bool = False,
            *_,
            **__
    ) -> Self:
        self = cls.__new__(cls)
        self.api = api

        # set a mock response with URL to load from
        id_ = api.wrangler.extract_ids(value)[0]
        self._response = {
            "href": api.wrangler.convert(
                id_, kind=RemoteObjectType.TRACK, type_in=RemoteIDType.ID, type_out=RemoteIDType.URL
            )
        }
        await self.reload(
            features=features, analysis=analysis, extend_album=extend_album, extend_artists=extend_artists
        )

        return self

    async def reload(
            self,
            features: bool = False,
            analysis: bool = False,
            extend_album: bool = False,
            extend_artists: bool = False,
            *_,
            **__
    ) -> None:
        self._check_for_api()

        # reload with enriched data
        response = await self.api.handler.get(self.url)
        if extend_album:
            await self.api.get_items(response["album"], kind=RemoteObjectType.ALBUM, extend=False)
        if extend_artists:
            await self.api.get_items(response["artists"], kind=RemoteObjectType.ARTIST)
        if features or analysis:
            await self.api.extend_tracks(response, features=features, analysis=analysis)

        self.__init__(response=response, api=self.api)


class SpotifyCollectionLoader[T: SpotifyObject](RemoteCollectionLoader[T], SpotifyObject, metaclass=ABCMeta):
    """Generic class for storing a collection of Spotify objects that can be loaded from an API response."""

    __slots__ = ()

    @classmethod
    def _get_item_kind(cls, api: SpotifyAPI) -> RemoteObjectType:
        """Returns the :py:class:`RemoteObjectType` for the items in this collection"""
        return api.collection_item_map[cls.kind]

    @classmethod
    @abstractmethod
    async def _get_items(
            cls, items: Collection[str] | MutableMapping[str, Any], api: SpotifyAPI
    ) -> list[dict[str, Any]]:
        """Call the ``api`` to get values for the given ``items`` URIs"""
        raise NotImplementedError

    @classmethod
    def _filter_items(cls, items: Iterable[T], response: Mapping[str, Any]) -> Iterable[T]:
        """
        Filter down and return only ``items`` associated with the given ``response``.
        The default implementation of this will just return the given ``items``.
        Override for object-specific filtering.
        """
        return items

    @classmethod
    async def _extend_response(cls, response: MutableMapping[str, Any], api: SpotifyAPI, *_, **__) -> bool:
        """
        Apply extensions to specific aspects of the given ``response``.
        Does nothing by default. Override to implement object-specific extensions.

        :return: True if checks should be skipped when initialising the object.
        """
        pass

    @classmethod
    def _merge_items_to_response(
            cls,
            items: Iterable[T | Mapping[str, Any]],
            response: Iterable[MutableMapping[str, Any]],
            skip: Collection[str] = ()
    ) -> tuple[set[str], set[str]]:
        """
        Find items in the ``response`` that match from the given ``items`` and replace them in the response.
        Optionally, provide a list of URIs to ``skip``.

        :return: Two sets of URIs for items that could and could not be found in response.
        """
        items_mapped: dict[str, T | Mapping[str, Any]] = {
            item.uri if isinstance(item, SpotifyObject) else item["uri"]: item for item in items
        }
        uri_matched: set[str] = set()
        uri_missing: set[str] = set()

        # find items in the response that match from the given items
        for source_item in response:
            if cls.kind == RemoteObjectType.PLAYLIST:
                source_item = source_item["track"]

            if source_item["uri"] in skip:
                continue
            elif source_item["uri"] in items_mapped:
                # replace the skeleton response with the response from the given items
                replacement_item = items_mapped.pop(source_item["uri"])
                if isinstance(replacement_item, SpotifyObject):
                    replacement_item = replacement_item.response

                source_item.clear()
                source_item |= replacement_item
                uri_matched.add(source_item["uri"])
            elif not source_item.get("is_local"):  # add to missing list
                uri_missing.add(source_item["uri"])

        return uri_matched, uri_missing

    @classmethod
    async def _load_new(cls, value: APIInputValueSingle[Self], api: SpotifyAPI, *args, **kwargs) -> Self:
        """
        Sets up a new object of the current class for the given ``value`` by calling ``__new__``
        and adding just enough attributes to the object to get :py:meth:`reload` to run.
        """
        # noinspection PyTypeChecker
        id_ = next(iter(api.wrangler.extract_ids(values=value, kind=cls.kind)))
        # noinspection PyTypeChecker
        url = api.wrangler.convert(id_, kind=cls.kind, type_in=RemoteIDType.ID, type_out=RemoteIDType.URL)

        self = cls.__new__(cls)
        self.api = api
        self._response = {"href": url}

        await self.reload(*args, **kwargs)
        return self

    @classmethod
    async def load(
            cls,
            value: APIInputValueSingle[Self],
            api: SpotifyAPI,
            items: Iterable[T] = (),
            leave_bar: bool = True,
            *args,
            **kwargs
    ) -> Self:
        item_kind = cls._get_item_kind(api=api)
        item_key = item_kind.name.lower() + "s"

        if isinstance(value, RemoteResponse):
            value = value.response

        # no items given, regenerate API response from the URL
        if any({not items, isinstance(value, Mapping) and api.items_key not in value.get(item_key, [])}):
            return await cls._load_new(value=value, api=api, *args, **kwargs)

        if isinstance(value, MutableMapping) and api.wrangler.get_item_type(value) == cls.kind:  # input is response
            response = deepcopy(value)
        else:  # load fresh response from the API
            response = await cls.api.get_items(value, kind=cls.kind)[0]

        # filter down input items to those that match the response
        items = cls._filter_items(items=items, response=response)
        matched, missing = cls._merge_items_to_response(items=items, response=response[item_key][api.items_key])

        if missing:
            items_missing = await cls._get_items(items=missing, api=api)
            cls._merge_items_to_response(items=items_missing, response=response[item_key][api.items_key], skip=matched)

        skip_checks = await cls._extend_response(response=response, api=api, *args, **kwargs)
        return cls(response=response, api=api, skip_checks=skip_checks)


class SpotifyPlaylist(SpotifyCollectionLoader[SpotifyTrack], RemotePlaylist[SpotifyTrack]):
    """
    Extracts key ``playlist`` data from a Spotify API JSON response.

    :param response: The Spotify API JSON response
    """

    __slots__ = ("_tracks",)
    __attributes_ignore__ = ("date_added",)

    @staticmethod
    def _validate_item_type(items: Any | Iterable[Any]) -> bool:
        if isinstance(items, Iterable):
            return all(isinstance(item, SpotifyTrack) for item in items)
        return isinstance(items, SpotifyTrack)

    @property
    def name(self):
        return self.response["name"]

    @name.setter
    def name(self, value: str):
        self.response["name"] = value

    @property
    def description(self):
        return self.response["description"]

    @description.setter
    def description(self, value: str | None):
        self.response["description"] = value

    @property
    def public(self):
        """Can other users access this playlist"""
        return self.response["public"]

    @public.setter
    def public(self, value: bool):
        self.response["public"] = value
        if value and self.collaborative:
            self.collaborative = False

    @property
    def collaborative(self):
        """Are other users allowed to modify this playlist"""
        return self.response["collaborative"]

    @collaborative.setter
    def collaborative(self, value: bool):
        if value and self.public:
            raise SpotifyCollectionError("You can only set collaborative to true on non-public playlists.")
        self.response["collaborative"] = value

    @property
    def followers(self):
        return int(self.response["followers"]["total"])

    @property
    def owner_name(self):
        return self.response["owner"]["display_name"]

    @property
    def owner_id(self):
        return self.response["owner"]["id"]

    @property
    def tracks(self):
        return self._tracks

    @property
    def track_total(self):
        return self.response["tracks"]["total"]

    @property
    def image_links(self):
        if not (images := self.response.get("images")):
            return {}

        images = {image["height"]: image["url"] for image in images}
        return {"cover_front": url for height, url in images.items() if height == max(images)}

    @property
    def date_created(self):
        """:py:class:`datetime` object representing when the first track was added to this playlist"""
        return min(self.date_added.values()) if self.date_added else None

    @property
    def date_modified(self):
        """:py:class:`datetime` object representing when a track was most recently added/removed"""
        return max(self.date_added.values()) if self.date_added else None

    @property
    def date_added(self):
        return {
            track["track"]["uri"]: datetime.strptime(track["added_at"], "%Y-%m-%dT%H:%M:%SZ")
            for track in self.response["tracks"]["items"]
        }

    def __init__(self, response: dict[str, Any], api: SpotifyAPI | None = None, skip_checks: bool = False):
        self._tracks: list[SpotifyTrack] | None = None
        super().__init__(response=response, api=api, skip_checks=skip_checks)

    def refresh(self, skip_checks: bool = False) -> None:
        self._tracks = [
            SpotifyTrack(track, api=self.api, skip_checks=skip_checks)
            for track in self._response.get("tracks", {}).get("items", [])
        ]
        if not skip_checks:
            self._check_total()

    @classmethod
    async def _get_items(
            cls, items: Collection[str] | MutableMapping[str, Any], api: SpotifyAPI
    ) -> list[dict[str, Any]]:
        return await api.get_tracks(items)

    @classmethod
    async def _extend_response(
            cls,
            response: MutableMapping[str, Any],
            api: SpotifyAPI,
            leave_bar: bool = True,
            extend_tracks: bool = False,
            extend_features: bool = False,
            *_,
            **__
    ) -> bool:
        item_kind = api.collection_item_map[cls.kind]

        if extend_tracks:
            # noinspection PyTypeChecker
            await api.extend_items(response, kind=cls.kind, key=item_kind, leave_bar=leave_bar)

        item_key = item_kind.name.lower() + "s"
        tracks = [item["track"] for item in response.get(item_key, {}).get(api.items_key, [])]
        if tracks and extend_features:
            await api.extend_tracks(tracks, limit=response[item_key]["limit"], features=True)

        return not extend_tracks

    async def reload(self, extend_tracks: bool = False, extend_features: bool = False, *_, **__) -> None:
        self._check_for_api()
        response = next(iter(await self.api.get_items(self.url, kind=RemoteObjectType.PLAYLIST, extend=False)))

        skip_checks = await self._extend_response(
            response=response, api=self.api, extend_tracks=extend_tracks, extend_features=extend_features
        )

        self.__init__(response=response, api=self.api, skip_checks=skip_checks)

    def _get_track_uris_from_api_response(self) -> list[str]:
        return [track["track"]["uri"] for track in self.response["tracks"].get("items", [])]


class SpotifyAlbum(RemoteAlbum[SpotifyTrack], SpotifyCollectionLoader[SpotifyTrack]):
    """
    Extracts key ``album`` data from a Spotify API JSON response.

    :param response: The Spotify API JSON response
    """

    __slots__ = ("_tracks", "_artists")

    @staticmethod
    def _validate_item_type(items: Any | Iterable[Any]) -> bool:
        if isinstance(items, Iterable):
            return all(isinstance(item, SpotifyTrack) for item in items)
        return isinstance(items, SpotifyTrack)

    @property
    def name(self):
        return self.response["name"]

    @property
    def tracks(self) -> list[SpotifyTrack]:
        return self._tracks

    @property
    def artists(self) -> list[SpotifyArtist]:
        return self._artists

    @property
    def artist(self):
        return self.tag_sep.join(artist["name"] for artist in self.response["artists"])

    @property
    def album_artist(self):
        return self.artist

    @property
    def track_total(self):
        return self.response["total_tracks"]

    @property
    def genres(self):
        if "genres" in self.response:
            return [g.title() for g in self.response["genres"]]
        main_artist_genres = self.response["artists"][0].get("genres", [])
        return [g.title() for g in main_artist_genres] if main_artist_genres else None

    @property
    def year(self):
        values = self.response["release_date"].split("-")
        return int(values[0]) if len(values) >= 1 and values[0].isdigit() else None

    @property
    def month(self):
        if self.response["release_date_precision"] in {"year"}:
            # release date is not precise to month, do not return month even if it exists in the value
            return
        values = self.response["release_date"].split("-")
        return int(values[1]) if len(values) >= 2 and values[1].isdigit() else None

    @property
    def day(self):
        if self.response["release_date_precision"] in {"year", "month"}:
            # release date is not precise to day, do not return day even if it exists in the value
            return
        values = self.response["release_date"].split("-")
        return int(values[2]) if len(values) >= 3 and values[2].isdigit() else None

    @property
    def compilation(self):
        return self.response["album_type"] == "compilation"

    @property
    def image_links(self):
        if not (images := self.response.get("images")):
            return {}

        images = {image["height"]: image["url"] for image in images}
        return {"cover_front": url for height, url in images.items() if height == max(images)}

    @property
    def rating(self):
        return self.response.get("popularity")

    def __init__(self, response: dict[str, Any], api: SpotifyAPI | None = None, skip_checks: bool = False):
        self._tracks: list[SpotifyTrack] | None = None
        self._artists: list[SpotifyArtist] | None = None

        if "album" in response and isinstance(response["album"], dict):
            # happens in 'user's saved ...' or playlist responses
            response = response["album"]

        super().__init__(response=response, api=api, skip_checks=skip_checks)

    def refresh(self, skip_checks: bool = False) -> None:
        album_only = copy(self.response)
        album_only.pop("tracks", None)
        for track in self.response.get("tracks", {}).get("items", []):
            track["album"] = album_only

        self._artists = [SpotifyArtist(artist, api=self.api) for artist in self._response.get("artists", {})]
        self._tracks = [
            SpotifyTrack(artist, api=self.api) for artist in self.response.get("tracks", {}).get("items", [])
        ]
        if not skip_checks:
            self._check_total()

        for track in self.tracks:
            track.disc_total = self.disc_total

    @classmethod
    async def _get_items(
            cls, items: Collection[str] | MutableMapping[str, Any], api: SpotifyAPI
    ) -> list[dict[str, Any]]:
        return await api.get_tracks(items)

    @classmethod
    def _filter_items(cls, items: Iterable[SpotifyTrack], response: Mapping[str, Any]) -> Iterable[SpotifyTrack]:
        """Filter down and return only ``items`` the items associated with the given ``response``"""
        return [item for item in items if item.response.get("album", {}).get("id") == response["id"]]

    @classmethod
    async def _extend_response(
            cls,
            response: MutableMapping[str, Any],
            api: SpotifyAPI,
            leave_bar: bool = True,
            extend_tracks: bool = False,
            extend_artists: bool = False,
            extend_features: bool = False,
            *_,
            **__
    ) -> bool:
        item_kind = api.collection_item_map[cls.kind]

        if extend_artists:
            await api.get_items(response["artists"], kind=RemoteObjectType.ARTIST)

        if extend_tracks:
            # noinspection PyTypeChecker
            await api.extend_items(response, kind=cls.kind, key=item_kind, leave_bar=leave_bar)

        item_key = item_kind.name.lower() + "s"
        tracks = response.get(item_key, {}).get(api.items_key)
        if tracks and extend_features:
            await api.extend_tracks(tracks, limit=response[item_key]["limit"], features=True)

        return not extend_tracks

    async def reload(
            self, extend_artists: bool = False, extend_tracks: bool = False, extend_features: bool = False, *_, **__
    ) -> None:
        self._check_for_api()
        response = next(iter(await self.api.get_items(self.url, kind=RemoteObjectType.ALBUM, extend=False)))

        skip_checks = await self._extend_response(
            response=response,
            api=self.api,
            extend_tracks=extend_tracks,
            extend_artists=extend_artists,
            extend_features=extend_features,
        )

        self.__init__(response=response, api=self.api, skip_checks=skip_checks)


class SpotifyArtist(RemoteArtist[SpotifyAlbum], SpotifyCollectionLoader[SpotifyAlbum]):
    """Extracts key ``artist`` data from a Spotify API JSON response."""

    __slots__ = ("_albums",)

    @staticmethod
    def _validate_item_type(items: Any | Iterable[Any]) -> bool:
        if isinstance(items, Iterable):
            return all(isinstance(item, SpotifyAlbum) for item in items)
        return isinstance(items, SpotifyAlbum)

    @property
    def name(self):
        return self.artist

    @property
    def items(self) -> list[SpotifyAlbum]:
        """The albums this artist is featured on"""
        return self._albums

    @property
    def artist(self):
        return self.response["name"]

    @property
    def albums(self) -> list[SpotifyAlbum]:
        album_frequencies = [
            (album, sum(artist.name == self.name for track in album for artist in track.artists))
            for album in self._albums
        ]
        return [album for album, _ in sorted(album_frequencies, key=lambda x: x[1])]

    @property
    def genres(self):
        return self.response.get("genres")

    @property
    def image_links(self):
        if not (images := self.response.get("images")):
            return {}

        images = {image["height"]: image["url"] for image in images}
        return {"cover_front": url for height, url in images.items() if height == max(images)}

    @property
    def rating(self):
        rating = self.response.get("popularity")
        return int(rating) if rating is not None else None

    @property
    def followers(self) -> int | None:
        """The total number of followers for this artist"""
        followers = self.response.get("followers", {}).get("total")
        return int(followers) if followers else None

    def __init__(self, response: dict[str, Any], api: SpotifyAPI | None = None, skip_checks: bool = False):
        self._albums: list[SpotifyAlbum] | None = None
        super().__init__(response=response, api=api, skip_checks=skip_checks)

    def refresh(self, skip_checks: bool = False) -> None:
        self._albums = [
            SpotifyAlbum(album, api=self.api, skip_checks=skip_checks)
            for album in self.response.get("albums", {}).get("items", [])
        ]

    @classmethod
    def _get_item_kind(cls, api: SpotifyAPI) -> RemoteObjectType:
        return RemoteObjectType.ALBUM

    @classmethod
    async def _get_items(
            cls, items: Collection[str] | MutableMapping[str, Any], api: SpotifyAPI
    ) -> list[dict[str, Any]]:
        return await api.get_items(items, extend=False)

    @classmethod
    def _filter_items(cls, items: Iterable[SpotifyAlbum], response: dict[str, Any]) -> Iterable[SpotifyAlbum]:
        """Filter down and return only ``items`` the items associated with the given ``response``"""
        return [
            item for item in items
            if any(artist["id"] == response["id"] for artist in item.response.get("artists", []))
        ]

    @classmethod
    async def _extend_response(
            cls,
            response: MutableMapping[str, Any],
            api: SpotifyAPI,
            leave_bar: bool = True,
            extend_albums: bool = False,
            extend_tracks: bool = False,
            extend_features: bool = False,
            *_,
            **__
    ) -> bool:
        item_kind = RemoteObjectType.ALBUM
        item_key = item_kind.name.lower() + "s"

        response_items = response.get(item_key, {})
        has_all_albums = item_key in response and len(response_items[api.items_key]) == response_items["total"]
        if extend_albums and not has_all_albums:
            await api.get_artist_albums(response, limit=response.get(item_key, {}).get("limit", 50))

        album_item_kind = api.collection_item_map[item_kind]
        album_item_key = album_item_kind.name.lower() + "s"
        albums = response.get(item_key, {}).get(api.items_key)

        if albums and extend_tracks:
            for album in albums:
                await api.extend_items(album[album_item_key], kind=item_kind, key=album_item_kind)

        if albums and extend_features:
            tracks = [track for album in albums for track in album[album_item_key]["items"]]
            await api.extend_tracks(tracks, limit=response[item_key]["limit"], features=True)

        return not extend_albums or not extend_tracks

    async def reload(
            self, extend_albums: bool = False, extend_tracks: bool = False, extend_features: bool = False, *_, **__
    ) -> None:
        self._check_for_api()
        response = await self.api.handler.get(url=self.url)

        skip_checks = await self._extend_response(
            response=response,
            api=self.api,
            extend_albums=extend_albums,
            extend_tracks=extend_tracks,
            extend_features=extend_features,
        )

        self.__init__(response=response, api=self.api, skip_checks=skip_checks)



"""
Core abstract classes for the :py:mod:`Spotify` module.

These define the foundations of any Spotify object or item.
"""
from abc import ABCMeta

from yarl import URL

from musify.libraries.remote.core.base import RemoteObject, RemoteItem
from musify.libraries.remote.core.exception import RemoteObjectTypeError, RemoteError
from musify.libraries.remote.core.types import RemoteObjectType
from musify.libraries.remote.spotify.api import SpotifyAPI


class SpotifyObject(RemoteObject[SpotifyAPI], metaclass=ABCMeta):
    """Generic base class for Spotify-stored objects. Extracts key data from a Spotify API JSON response."""

    __slots__ = ()

    @property
    def id(self):
        return self.response["id"]

    @property
    def uri(self):
        return self.response["uri"]

    @property
    def has_uri(self):
        return not self.response.get("is_local", False)

    @property
    def url(self):
        return URL(self.response["href"])

    @property
    def url_ext(self):
        url = self.response["external_urls"].get("spotify")
        return URL(url) if url else None

    def _check_type(self) -> None:
        """
        Checks the given response is compatible with this object type, raises an exception if not.

        :raise RemoteObjectTypeError: When the response type is not compatible with this object.
        """
        required_keys = {"id", "uri", "href", "external_urls"}
        for key in required_keys:
            if key not in self.response:
                raise RemoteError(f"Response does not contain all required keys: {required_keys}")

        kind = self.__class__.__name__.removeprefix("Spotify").lower()
        if self.response.get("type") != kind:
            kind = RemoteObjectType.from_name(kind)[0]
            raise RemoteObjectTypeError("Response type invalid", kind=kind, value=self.response.get("type"))


class SpotifyItem(SpotifyObject, RemoteItem, metaclass=ABCMeta):
    """Generic base class for Spotify-stored items. Extracts key data from a Spotify API JSON response."""

    __slots__ = ()



"""
Base implementation for the functionality of a local playlist.
"""
from abc import ABCMeta, abstractmethod
from collections.abc import Collection, Generator
from pathlib import Path
from typing import Any, Self

from musify.base import Result
from musify.file.base import File
from musify.file.path_mapper import PathMapper
from musify.libraries.core.object import Playlist
from musify.libraries.local.collection import LocalCollection
from musify.libraries.local.track import LocalTrack
from musify.libraries.remote.core.wrangle import RemoteDataWrangler
from musify.processors.base import Filter
from musify.processors.limit import ItemLimiter
from musify.processors.sort import ItemSorter


class LocalPlaylist[T: Filter[LocalTrack]](File, LocalCollection[LocalTrack], Playlist[LocalTrack], metaclass=ABCMeta):
    """
    Generic class for loading and manipulating local playlists.

    :param path: Absolute path of the playlist.
    :param matcher: :py:class:`Filter` object to use for matching tracks.
    :param limiter: :py:class:`ItemLimiter` object to use for limiting the number of tracks matched.
    :param sorter: :py:class:`ItemSorter` object to use for sorting the final track list.
    :param path_mapper: Optionally, provide a :py:class:`PathMapper` for paths stored in the playlist file.
        Useful if the playlist file contains relative paths and/or paths for other systems that need to be
        mapped to absolute, system-specific paths to be loaded and back again when saved.
    :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object for processing URIs on tracks.
        If given, the wrangler can be used when calling __get_item__ to get an item from the collection from its URI.
        The wrangler is also used when loading tracks to allow them to process URI tags.
        For more info on this, see :py:class:`LocalTrack`.
    """

    __slots__ = (
        "_path",
        "matcher",
        "limiter",
        "sorter",
        "path_mapper",
        "_tracks",
        "_original",
    )
    __attributes_classes__ = (Playlist, LocalCollection, File)

    @property
    def name(self) -> str:
        """The name of this playlist, always the same as the filename."""
        return self.filename

    @name.setter
    def name(self, value: str):
        self._path = self.path.with_stem(value).with_suffix(self.ext)

    @property
    def tracks(self) -> list[LocalTrack]:
        return self._tracks

    @tracks.setter
    def tracks(self, value: list[LocalTrack]):
        self._tracks = value

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value: str | Path):
        self._path = Path(value)

    def __init__(
            self,
            path: str | Path,
            matcher: T | None = None,
            limiter: ItemLimiter | None = None,
            sorter: ItemSorter | None = None,
            path_mapper: PathMapper = PathMapper(),
            remote_wrangler: RemoteDataWrangler = None,
    ):
        super().__init__(remote_wrangler=remote_wrangler)

        self._path: Path = Path(path)
        self._validate_type(self._path)

        #: :py:class:`Filter` object to use for matching tracks.
        self.matcher = matcher
        #: :py:class:`ItemLimiter` object to use for limiting the number of tracks matched.
        self.limiter = limiter
        #: :py:class:`ItemSorter` object to use for sorting the final track list.
        self.sorter = sorter
        #: Maps paths stored in the playlist file.
        self.path_mapper = path_mapper

        self._tracks: list[LocalTrack] = []
        self._original: list[LocalTrack] = []

    def __await__(self) -> Generator[Any, None, Self]:
        return self.load().__await__()

    def _match(self, tracks: Collection[LocalTrack] = (), reference: LocalTrack | None = None) -> None:
        if self.matcher is None or not tracks:
            return

        if not self.matcher.ready:  # just return the tracks given if matcher has no settings applied
            self.tracks: list[LocalTrack] = list(tracks)
        else:  # run matcher
            self.tracks: list[LocalTrack] = list(self.matcher(values=tracks, reference=reference))

    def _limit(self, ignore: Collection[str | LocalTrack]) -> None:
        if self.limiter is not None and self.tracks is not None:
            track_path_map = {str(track.path): track for track in self.tracks}
            ignore: set[LocalTrack] = {i if isinstance(i, LocalTrack) else track_path_map.get(i) for i in ignore}
            self.limiter(items=self.tracks, ignore=ignore)

    def _sort(self) -> None:
        if self.sorter is not None and self.tracks is not None:
            self.sorter(items=self.tracks)

    @abstractmethod
    async def load(self, tracks: Collection[LocalTrack] = ()) -> Self:
        """
        Read the playlist file and update the tracks in this playlist instance.

        :param tracks: Available Tracks to search through for matches.
        :return: Self
        """
        raise NotImplementedError

    @abstractmethod
    async def save(self, dry_run: bool = True, *args, **kwargs) -> Result:
        """
        Write the tracks in this Playlist and its settings (if applicable) to file.

        :param dry_run: Run function, but do not modify the file on the disk.
        :return: :py:class:`Result` object with stats on the changes to the playlist.
        """
        raise NotImplementedError



"""
Configuration relating to the :py:mod:`Spotify` module.

This configuration can be used to inject dependencies into dependencies throughout the module.
"""
from dataclasses import dataclass

from musify.libraries.remote.core.factory import RemoteObjectFactory
from musify.libraries.remote.spotify.object import SpotifyPlaylist, SpotifyTrack, SpotifyAlbum, SpotifyArtist


@dataclass
class SpotifyObjectFactory(RemoteObjectFactory):
    playlist: type[SpotifyPlaylist] = SpotifyPlaylist
    track: type[SpotifyTrack] = SpotifyTrack
    album: type[SpotifyAlbum] = SpotifyAlbum
    artist: type[SpotifyArtist] = SpotifyArtist



"""
All classes and operations relating to the logger objects used throughout the entire package.
"""
import asyncio
import logging
import logging.config
import logging.handlers
import os
import sys
from collections.abc import Iterable, Awaitable
from pathlib import Path
from typing import Any

try:
    from tqdm.auto import tqdm
except ImportError:
    tqdm = None

type ProgressBarType[T] = Iterable[T] | tqdm if tqdm is not None else Iterable[T]


INFO_EXTRA = logging.INFO - 1
logging.addLevelName(INFO_EXTRA, "INFO_EXTRA")
logging.INFO_EXTRA = INFO_EXTRA

REPORT = logging.INFO - 3
logging.addLevelName(REPORT, "REPORT")
logging.REPORT = REPORT

STAT = logging.DEBUG + 3
logging.addLevelName(STAT, "STAT")
logging.STAT = STAT


class MusifyLogger(logging.Logger):
    """The logger for all logging operations in Musify."""

    __slots__ = ()

    #: When true, never print a new line in the console when :py:meth:`print()` is called
    compact: bool = False
    #: When true, all bars returned by :py:meth:`get_progress_bar()` will be disabled by default
    disable_bars: bool = False
    #: All currently active progress bars
    _bars: list[tqdm] = []

    @property
    def file_paths(self) -> list[Path]:
        """Get a list of the paths of all file handlers for this logger"""
        def extract_paths(lggr: logging.Logger) -> None:
            """Extract file path from the handlers of the given ``lggr``"""
            for handler in lggr.handlers:
                if isinstance(handler, logging.FileHandler) and handler.baseFilename not in paths:
                    paths.append(Path(handler.baseFilename))

        paths = []
        logger = self
        extract_paths(logger)
        while logger.propagate and logger.parent:
            logger = logger.parent
            extract_paths(logger)
        return paths

    @property
    def stdout_handlers(self) -> set[logging.StreamHandler]:
        """Get a list of all :py:class:`logging.StreamHandler` handlers that log to stdout"""
        console_handlers = set()
        for handler in self.handlers + list(logging.getHandlerNames()):
            if isinstance(handler, str):
                handler = logging.getHandlerByName(handler)
            if isinstance(handler, logging.StreamHandler) and handler.stream == sys.stdout:
                console_handlers.add(handler)

        return console_handlers

    def info_extra(self, msg, *args, **kwargs) -> None:
        """Log 'msg % args' with severity 'INFO_EXTRA'."""
        if self.isEnabledFor(INFO_EXTRA):
            self._log(INFO_EXTRA, msg, args, **kwargs)

    def report(self, msg, *args, **kwargs) -> None:
        """Log 'msg % args' with severity 'REPORT'."""
        if self.isEnabledFor(REPORT):
            self._log(REPORT, msg, args, **kwargs)

    def stat(self, msg, *args, **kwargs) -> None:
        """Log 'msg % args' with severity 'STAT'."""
        if self.isEnabledFor(STAT):
            self._log(STAT, msg, args, **kwargs)

    def print_line(self, level: int = logging.CRITICAL + 1) -> None:
        """Print a new line only when DEBUG < ``logger level`` <= ``level`` for all console handlers"""
        if not self.compact:
            if self.stdout_handlers and any(logging.DEBUG < h.level <= level for h in self.stdout_handlers):
                print()

    def print_message(self, *values, sep=' ', end='\n') -> None:
        """
        Wrapper for print. Logs the given ``values`` to the INFO setting.
        If there are no stdout handlers with severity <= INFO, also print this to the terminal.
        This ensures the user sees the ``values`` always.
        """
        message = sep.join(values)
        if message:
            self.info(message)

        if not values or not self.stdout_handlers or all(h.level > logging.INFO for h in self.stdout_handlers):
            print(*values, sep=sep, end=end)

    def get_synchronous_iterator[T: Any](
            self, iterable: Iterable[T] | None = None, total: T | int | None = None, **kwargs
    ) -> ProgressBarType[T]:
        """
        Return an appropriately configured tqdm progress bar if installed.
        If not, return either the given ``iterable`` if given or simply ``range(total)``.
        For tqdm kwargs, see :py:class:`tqdm`
        """
        if tqdm is None:
            return iter(iterable) if iterable is not None else range(total)

        bar = tqdm(iterable=iterable, **self._get_tqdm_kwargs(total=total, **kwargs))
        self._bars.append(bar)
        return bar

    def get_asynchronous_iterator[T](self, tasks: Iterable[Awaitable[T]], **kwargs) -> Awaitable[list[T]]:
        """
        Return an appropriately configured asynchronous tqdm progress bar if installed.
        If not, gather the given awaitable objects from ``tasks`` and return a coroutine.

        Note that tqdm does not preserve the order of the input awaitables and will return results in a random order.
        For tqdm kwargs, see :py:class:`tqdm`
        """
        if tqdm is None:
            return asyncio.gather(*tasks)

        return tqdm.gather(*tasks, **self._get_tqdm_kwargs(**kwargs))

    def _get_tqdm_kwargs(self, **kwargs) -> dict[str, Any]:
        # noinspection SpellCheckingInspection
        preset_keys = ("leave", "disable", "file", "ncols", "colour", "smoothing")

        try:
            cols = os.get_terminal_size().columns
        except OSError:
            cols = 120

        # adjust kwargs to defaults if needed
        kwargs["position"] = self._get_tqdm_param_position(**kwargs)
        return dict(
            leave=self._get_tqdm_param_leave(**kwargs),
            disable=self.disable_bars or kwargs.get("disable", False),
            file=sys.stdout,
            ncols=cols,
            colour=kwargs.get("colour", "blue"),
            smoothing=0.1,
            **{k: v for k, v in kwargs.items() if k not in preset_keys}
        )

    def _get_tqdm_param_position(self, position: int = None, **__) -> int | None:
        if position is not None:
            return position

        # clear closed bars
        self._bars = [bar for bar in self._bars if bar.n < bar.total]
        if self._bars:
            return abs(min(bar.pos for bar in self._bars)) + 1

    def _get_tqdm_param_leave(self, position: int | None, leave: bool = None, **__) -> bool:
        if leave is not None:
            return leave

        return all([
            bool(self.stdout_handlers) or (h.level > logging.DEBUG for h in self.stdout_handlers),
            position is None or position == 0
        ])

    def __copy__(self):
        """Do not copy logger"""
        return self

    def __deepcopy__(self, _: dict = None):
        """Do not copy logger"""
        return self


logging.setLoggerClass(MusifyLogger)



"""
Exceptions relating to Spotify operations.
"""
from musify.libraries.remote.core.exception import RemoteError


class SpotifyError(RemoteError):
    """
    Exception raised for Spotify ID errors.

    :param message: Explanation of the error.
    """
    def __init__(self, message: str | None = None):
        self.message = message
        super().__init__(message)


class SpotifyItemError(SpotifyError):
    """
    Exception raised for Spotify item errors.

    :param message: Explanation of the error.
    :param kind: The item type related to the error.
    """
    def __init__(self, message: str | None = None, kind: str | None = None):
        self.message = message
        self.kind = kind
        formatted = f"{kind} | {message}" if kind else message
        super().__init__(formatted)


class SpotifyCollectionError(SpotifyError):
    """
    Exception raised for Spotify collection errors.

    :param message: Explanation of the error.
    :param kind: The collection type related to the error.
    """
    def __init__(self, message: str | None = None, kind: str | None = None):
        self.message = message
        self.kind = kind
        formatted = f"{kind} | {message}" if kind else message
        super().__init__(formatted)



"""
Processor that helps user download songs from collections based on given configuration.
"""
import re
from collections.abc import Iterable, Collection
from itertools import batched
from typing import Any
from urllib.parse import quote
from webbrowser import open as webopen

from musify.base import MusifyItem, MusifyObject
from musify.exception import MusifyEnumError
from musify.field import Field, Fields
from musify.libraries.core.collection import MusifyCollection
from musify.processors.base import InputProcessor
from musify.types import UnitIterable
from musify.utils import to_collection


class ItemDownloadHelper(InputProcessor):
    """
    Runs operations for helping the user to download items from given collections.

    :param urls: The template URLs for websites to open queries for.
        The given sites should contain exactly 1 '{}' placeholder into which the processor can place
        a query for the item being searched. e.g. *bandcamp.com/search?q={}&item_type=t*
    :param fields: The default fields to take from an item for use as the query string when initially opening sites.
    :param interval: The number of items to open sites for before pausing for user input.
    """

    __slots__ = ("urls", "fields", "interval")

    def __init__(self, urls: UnitIterable[str] = (), fields: UnitIterable[Field] = Fields.ALL, interval: int = 1):
        super().__init__()

        self.urls: list[str] = to_collection(urls, list)
        if fields == Fields.ALL or Fields.ALL in to_collection(fields, set):
            fields = Fields.all()
        self.fields: list[Field] = to_collection(fields, list)
        self.interval = interval

    def __call__(self, *args, **kwargs) -> None:
        return self.open_sites(*args, **kwargs)

    def open_sites(self, collections: UnitIterable[MusifyCollection]) -> None:
        """
        Run the download helper.

        Opens the formatted ``urls`` for each item in all given collections in the user's browser.
        """
        if isinstance(collections, MusifyCollection):
            items = collections.items
        else:
            items = [item for coll in collections for item in coll]

        pages_total = (len(items) // self.interval) + (len(items) % self.interval > 0)
        for page, items in enumerate(batched(items, self.interval), 1):
            not_queried = self._open_sites_for_items(items=items, fields=self.fields)
            self._pause(items=items, not_queried=not_queried, page=page, total=pages_total)

    def _open_sites_for_items(self, items: Iterable[MusifyItem], fields: Iterable[Field]) -> list[MusifyItem]:
        not_queried = []
        for item in items:
            queried = self._open_sites_for_item(item=item, fields=fields)
            if not queried:
                not_queried.append(item)

        return not_queried

    def _open_sites_for_item(self, item: MusifyItem, fields: Iterable[Field]) -> bool:
        query_parts = []
        for field in fields:
            field_name = field.name.lower()
            if (value := getattr(item, field_name, None)) is None:
                continue

            if isinstance((value_many := getattr(item, field_name + "s", None)), (list, tuple)):
                value = next(iter(value_many))

            if isinstance(value, MusifyObject):
                value = value.name
            elif isinstance(value, (tuple, set, list, dict)):
                value = " ".join(v.name if isinstance(v, MusifyObject) else v for v in value)

            query_parts.append(str(value))

        query = quote(" ".join(query_parts))
        if not query:
            self.logger.debug(f"Could not get query for item: {item.name}")
            return False

        self.logger.debug(f"Opening {len(self.urls)} URLs with query: {query}")
        for url in self.urls:
            webopen(url.format(query))

        return True

    def _pause(self, items: Iterable[MusifyItem], not_queried: Collection[MusifyItem], page: int, total: int):
        opened = len(self.urls) * (self.interval - len(not_queried))
        not_opened = f" - Could not open sites for {len(not_queried)} items. " if not_queried else ". "
        valid_fields = [
            field.name.lower() for field in Fields.all()
            if any(
                hasattr(item, field.name.lower()) and getattr(item, field.name.lower()) is not None for item in items
            )
        ]

        header = [
            f"\t\33[1;94mOpened {opened} sites" + not_opened + "You may now search for and download the items. \33[0m"
        ]
        options = {
            "<Return/Enter>": "Once you are finished with this batch, continue on to the next batch",
            "r": "Re-open all sites for the current batch of items",
            "<Fields>":
                "Re-open all sites for the current batch of items using the input list of fields, "
                "each separated by a space e.g. title artist album",
        }

        if not_queried:
            options["n <Fields>"] = (
                f"Same as above, but only open sites for the {len(not_queried)} items "
                "which sites could not be opened for"
            )
        options["h"] = "Show this dialogue again"

        current_input = "START"
        help_text = self._format_help_text(options=options, header=header)
        help_text += f"\n\t\33[90mValid fields for this batch: {" ".join(valid_fields)}\33[0m\n"

        self.logger.print_message("\n" + help_text)
        while current_input != '':
            current_input = self._get_user_input(f"Enter ({page}/{total})")

            if current_input.casefold() == "h":  # print help text
                self.logger.print_message("\n" + help_text)

            elif current_input.casefold() == "r":
                self._open_sites_for_items(items=items, fields=self.fields)

            elif current_input != "":
                try:
                    fields = [
                        enum for field in re.sub(r"^n ", "", current_input).split(" ")
                        for enum in Fields.from_name(field)
                    ]
                except MusifyEnumError:
                    self.logger.warning(
                        "Some fields were not recognised. "
                        f"Please only use one of the following fields: {", ".join(valid_fields)}"
                    )
                    continue
                self._open_sites_for_items(
                    items=not_queried if current_input.startswith("n ") else items, fields=fields
                )

    def as_dict(self) -> dict[str, Any]:
        return {
            "urls": self.urls,
            "fields": [field.name.lower() for field in self.fields],
            "interval": self.interval,
        }



"""
The base processor definition for reading/manipulating tag data in an audio file.
"""
from abc import ABCMeta

import mutagen

from musify.field import TagMap
from musify.libraries.local.track.field import LocalTrackField
from musify.libraries.remote.core.wrangle import RemoteDataWrangler


class TagProcessor[T: mutagen.FileType](metaclass=ABCMeta):
    """
    Base tag processor for reading/writing of tag metadata to a file.

    :param file: The loaded Mutagen object of the file to extract tags from.
    :param tag_map: The map of tag names to tag IDs for the given file type.
    :param remote_wrangler: Optionally, provide a :py:class:`RemoteDataWrangler` object for processing URIs.
        This object will be used to check for and validate a URI tag on the file.
        The tag that is used for reading and writing is set by the ``uri_tag`` class attribute.
        If no ``remote_wrangler`` is given, no URI processing will occur.
    """

    __slots__ = ("file", "tag_map", "remote_wrangler")

    #: The tag field to use as the URI tag in the file's metadata
    uri_tag: LocalTrackField = LocalTrackField.COMMENTS
    #: The separator to use when representing separated tag values as a combined string.
    #: Used when some number type tag values come as a combined string i.e. track number/track total
    num_sep: str = "/"

    @property
    def remote_source(self) -> str | None:
        """The name of the remote service for which this TagProcessor can process remote data."""
        return self.remote_wrangler.source if self.remote_wrangler else None

    @property
    def unavailable_uri_dummy(self) -> str | None:
        """
        The value to use as a URI for an item which does not have an associated remote object.
        An item that has this URI value will be excluded from most remote logic.
        If no remote_wrangler has been assigned to this object, return None.
        """
        return self.remote_wrangler.unavailable_uri_dummy if self.remote_wrangler else None

    def __init__(self, file: T, tag_map: TagMap, remote_wrangler: RemoteDataWrangler | None = None):
        self.file = file
        self.tag_map = tag_map
        self.remote_wrangler = remote_wrangler
